"""mactypes -- Defines user-friendly wrapper classes for Mac OS datatypes that don't have a suitable Python equivalent.

- File objects encompass AEDescs of typeFSRef/typeFSSpec/typeFileURL to save user from having to deal with them directly. File objects refer to filesystem locations which may or may not already exist.

- Alias objects wrap AEDescs of typeAlias. Aliases refer to filesystem objects and can track them as they're moved around the disk or renamed.

Both classes provide a variety of constructors and read-only properties for getting raw objects in and out. Objects are comparable and nominally hashable.

- Units objects represent units of measurement. Default unit types are defined in aem; clients can add additional definitions to aem Codecs/appscript AppData objects as needed.

"""

from os.path import abspath

from ae import newdesc, convertpathtourl, converturltopath, MacOSError
import kae

kCFURLPOSIXPathStyle = 0
kCFURLHFSPathStyle = 1
kCFURLWindowsPathStyle = 2

__all__ = ['Alias', 'File', 'Units']


######################################################################
# PRIVATE
######################################################################
# Constants

class _kNoPath: pass

class _Base(object):
	def __eq__(self, val):
		return self is val or (self.__class__ == val.__class__ and self.url == val.url)
	
	def __ne__(self, val):
		return not self == val
	
	def __hash__(self):
		return hash(self.__class__)


######################################################################
# PUBLIC
######################################################################

class Alias(_Base):
	"""A persistent reference to a filesystem object."""
	
	# Constructors
	
	def __init__(self, path):
		"""Make Alias object from POSIX path."""
		if path is _kNoPath:
			self._desc = None
		else:
			urldesc = newdesc(kae.typeFileURL, 
					convertpathtourl(abspath(path), kCFURLPOSIXPathStyle))
			try:
				self._desc = urldesc.coerce(kae.typeAlias)
			except MacOSError, err:
				if err.args[0] == -1700:
					raise ValueError("Can't make mactypes.Alias as file doesn't exist: %r" % path)
				else:
					raise
		
	def makewithhfspath(klass, path):
		return klass.makewithurl(convertpathtourl(path, kCFURLHFSPathStyle))
	makewithhfspath = classmethod(makewithhfspath)
	
	def makewithurl(klass, url):
		"""Make File object from file URL."""
		obj = klass(_kNoPath)
		obj._desc = newdesc(kae.typeFileURL, url).coerce(kae.typeAlias)
		return obj
	makewithurl = classmethod(makewithurl)
	
	def makewithdesc(klass, desc):
		"""Make Alias object from aem.ae.AEDesc of typeAlias (typeFSS/typeFSRef/typeFileURL are also allowed).
		"""
		if desc.type != kae.typeAlias:
			desc = desc.coerce(kae.typeAlias)
		obj = klass(_kNoPath)
		obj._desc = desc
		return obj
	makewithdesc = classmethod(makewithdesc)
	
	# Instance methods
	
	def __repr__(self):
		return 'mactypes.Alias(%r)' % unicode(self.path)
	
	# Properties
	
	path = property(lambda self: converturltopath(self.url, kCFURLPOSIXPathStyle), doc="Get as POSIX path.")
	
	hfspath = property(lambda self: converturltopath(self.url, kCFURLHFSPathStyle), doc="Get as HFS path.")
	
	url = property(lambda self: self._desc.coerce(kae.typeFileURL).data, doc="Get as file URL.")
	
	file = property(lambda self: File.makewithdesc(self._desc), doc="Get as mactypes.File.")
	
	alias = property(lambda self: self, doc="Get as mactypes.Alias (i.e. itself).")
	
	desc = property(lambda self: self._desc, doc="Get as aem.ae.AEDesc.")



class File(_Base):
	"""A reference to a filesystem location."""
	
	# Constructors
	
	def __init__(self, path):
		"""Make File object from POSIX path."""
		if path is not _kNoPath:
			if not isinstance(path, unicode):
				path = unicode(path)
			self._path = abspath(path)
			self._url = convertpathtourl(self._path, kCFURLPOSIXPathStyle)
			self._desc = newdesc(kae.typeFileURL, self._url)
	
	def makewithhfspath(klass, path):
		return klass.makewithurl(convertpathtourl(path, kCFURLHFSPathStyle))
	makewithhfspath = classmethod(makewithhfspath)
	
	def makewithurl(klass, url):
		"""Make File object from file URL."""
		obj = klass(_kNoPath)
		obj._desc = newdesc(kae.typeFileURL, url)
		obj._url = url
		obj._path = converturltopath(url, kCFURLPOSIXPathStyle)
		return obj
	makewithurl = classmethod(makewithurl)
		
	def makewithdesc(klass, desc):
		"""Make File object from aem.ae.AEDesc of typeFSS/typeFSRef/typeFileURL.
			Note: behaviour for other descriptor types is undefined: typeAlias will cause problems, others will probably fail.
		"""
		obj = klass(_kNoPath)
		obj._path = None
		obj._url = None
		if desc.type in [kae.typeFSS, kae.typeFSRef, kae.typeFileURL]:
			obj._desc = desc
		else:
			obj._desc = desc.coerce(kae.typeFileURL)
		return obj
	makewithdesc = classmethod(makewithdesc)
	
	# Instance methods
	
	def __repr__(self):
		return 'mactypes.File(%r)' % self.path
	
	# Properties
	
	def path(self):
		if self._path is None:
			self._path = converturltopath(self.url, kCFURLPOSIXPathStyle)
		return self._path
	path = property(path, doc="Get as POSIX path.")
	
	hfspath = property(lambda self: converturltopath(self.url, kCFURLHFSPathStyle), doc="Get as HFS path.")
	
	def url(self):
		if self._url is None:
			if self._desc.type == kae.typeFileURL:
				self._url = self._desc.data
			else:
				self._url = self._desc.coerce(kae.typeFileURL).data
		return self._url
	url = property(url, doc="Get as file URL.")
	
	file = property(lambda self: File(self.path), doc="Get as mactypes.File.")
	
	alias = property(lambda self: Alias.makewithdesc(self.desc), doc="Get as mactypes.Alias.")
	
	def desc(self):
		if self._desc is None:
			self._desc = newdesc(kae.typeFileURL, self.url)
		return self._desc
	desc = property(desc, doc="Get as aem.ae.AEDesc.")



#######

class Units(object):
	"""Represents a measurement; e.g. 3 inches, 98.5 degrees Fahrenheit.
	
	The AEM defines a standard set of unit types; some applications may define additional types for their own use. This wrapper stores the raw unit type and value data; aem/appscript Codecs objects will convert this to/from an AEDesc, or raise an error if the unit type is unrecognised.
	"""
	
	def __init__(self, value, type):
		"""
			value : int | float -- the unit value, e.g. 3
			type : str -- the unit type name, e.g. 'inches'
		"""
		self._value = value
		self._type = type
	
	value = property(lambda self: self._value, doc="Get unit value, e.g. 3")
	type = property(lambda self: self._type, doc="Get unit type, e.g. 'inches'")
	
	def __eq__(self, val):
		return self is val or (self.__class__ == val.__class__ 
				and self._value == val.value and self._type == val.type)
	
	def __ne__(self, val):
		return not self == val
	
	def __hash__(self):
		return hash((self.value, self.type))
	
	def __repr__(self):
		return 'mactypes.Units(%r, %r)' % (self.value, self.type)
	
	def __str__(self):
		return '%r %s' % (self.value, self.type.replace('_', ' '))
	
	def __int__(self):
		return int(self.value)
	
	def __float__(self):
		return float(self.value)





"""keywordwrapper -- generic wrapper for application-specific type and enum names. """

# The Keyword class provides a generic wrapper for class, enum, property and type names.
# Users don't instantiate this class directly; instead, the syntactic sugar layer allows keywords
# to be created by referring to the exported 'k' variable; e.g. k.document, k.ask, k.name, k.String.

class Keyword(object):
	"""A class/property/enumerator/type name."""
	
	def __init__(self, name): 
		self.AS_name = name
	
	def __repr__(self): 
		return 'k.%s' % self.AS_name
	
	def __hash__(self): 
		return hash(self.AS_name)
	
	def __eq__(self, val):
		return val.__class__ == self.__class__ and val.AS_name == self.AS_name
	
	def __ne__(self, val):
		return not self.__eq__(val)
	
	def __nonzero__(self):
		return self.AS_name != 'missing_value'
	
	name = property(lambda self:self.AS_name)


class _KeywordShim(object):
	""" Infinite namespace 'containing' all possible class/property/enumerator/type names. """
	
	def __getattr__(self, name): 
		if name.startswith('__') and name.endswith('__'):
			return object.__getattr__(self, name)
		else:
			return Keyword(name)
	
	def __repr__(self):
		return 'k'	


k = _KeywordShim()




"""typewrappers -- wrapper classes for AE type and enumeration codes """

######################################################################
# PUBLIC
######################################################################

class AETypeBase(object): # base class; exposed for typechecking purposes
	
	def __init__(self, code):
		# Check arg is a 4-byte code (while ae.newdesc() verifies descriptor type codes okay, it doesn't verify data size so wouldn't catch bad values at packing time):
		if not isinstance(code, basestring):
			raise TypeError('invalid code (not a str object): %r' % code)
		elif len(code) != 4:
			raise ValueError('invalid code (not four bytes long): %r' % code)
		self._code = code
	
	code = property(lambda self:self._code)
	
	def __hash__(self): 
		return hash(self._code)
	
	def __eq__(self, val):
		return val.__class__ == self.__class__ and val.code == self._code
	
	def __ne__(self, val):
		return not self == val
		
	def __repr__(self):
		s = ''
		for c in self._code:
			if 33 < ord(c) < 127 and c != "'":
				s += c
			else:
				s += '\\x%2.2x' % ord(c)
		return "aem.%s('%s')" % (self.__class__.__name__, s)


class AEType(AETypeBase):
	"""An AE type."""


class AEEnum(AETypeBase):
	"""An AE enumeration."""


class AEProp(AETypeBase):
	"""An AE property code."""


class AEKey(AETypeBase):
	"""An AE keyword."""




print("import: 'aem'")
import aem

print("import: 'appscript'")
import appscript

print("import: 'mactypes'")
import mactypes

print("import: 'osax'")
import osax

print("import: 'aem.ae'")
import aem.ae




"""connect - Creates Apple event descriptor records of typeProcessSerialNumber and typeApplicationURL, used to specify the target application in send.Event() constructor. """
import struct
from time import sleep

import ae, kae

from aemsend import Event

__all__ = ['launchapp', 'processexistsforpath', 'processexistsforpid', 'processexistsforurl', 
		'processexistsfordesc', 'currentapp', 'localapp', 'remoteapp', 'CantLaunchApplicationError']

######################################################################
# PRIVATE
######################################################################

kLSLaunchDefaults = 0x00000001
kLSLaunchAndPrint = 0x00000002
kLSLaunchReserved2 = 0x00000004
kLSLaunchReserved3 = 0x00000008
kLSLaunchReserved4 = 0x00000010
kLSLaunchReserved5 = 0x00000020
kLSLaunchAndDisplayErrors = 0x00000040
kLSLaunchInhibitBGOnly = 0x00000080
kLSLaunchDontAddToRecents = 0x00000100
kLSLaunchDontSwitch = 0x00000200
kLSLaunchNoParams = 0x00000800
kLSLaunchAsync = 0x00010000
kLSLaunchStartClassic = 0x00020000
kLSLaunchInClassic = 0x00040000
kLSLaunchNewInstance = 0x00080000
kLSLaunchAndHide = 0x00100000
kLSLaunchAndHideOthers = 0x00200000
kLSLaunchHasUntrustedContents = 0x00400000


_kNoProcess = 0
_kCurrentProcess = 2

_nulladdressdesc = ae.newdesc(kae.typeProcessSerialNumber, struct.pack('II', 0, _kNoProcess)) # ae.newappleevent complains if you pass None as address, so we give it one to throw away

_launchevent = Event(_nulladdressdesc, 'ascrnoop').AEM_event
_runevent = Event(_nulladdressdesc, 'aevtoapp').AEM_event

#######

def _launchapplication(path, event, newinstance=False, hide=False):
	flags = kLSLaunchNoParams | kLSLaunchStartClassic | kLSLaunchDontSwitch
	if newinstance:
		flags |= kLSLaunchNewInstance
	if hide:
		flags |= kLSLaunchAndHide
	try:
		return ae.launchapplication(path, event, flags)
	except ae.MacOSError, err:
		raise CantLaunchApplicationError(err.args[0], path)


######################################################################
# PUBLIC
######################################################################


class CantLaunchApplicationError(Exception):
	
	_lserrors = {
		# following taken from <http://developer.apple.com/documentation/Carbon/Reference/LaunchServicesReference>:
		-10660: "The application cannot be run because it is inside a Trash folder.",
		-10810: "An unknown error has occurred.",
		-10811: "The item to be registered is not an application.",
		-10813: "Data of the desired type is not available (for example, there is no kind string).",
		-10814: "No application in the Launch Services database matches the input criteria.",
		-10817: "Data is structured improperly (for example, an item's information property list is malformed).",
		-10818: "A launch of the application is already in progress.",
		-10822: "There is a problem communicating with the server process that maintains the Launch Services database.",
		-10823: "The filename extension to be hidden cannot be hidden.",
		-10825: "The application to be launched cannot run on the current Mac OS version.",
		-10826: "The user does not have permission to launch the application (on a managed network).",
		-10827: "The executable file is missing or has an unusable format.",
		-10828: "The Classic emulation environment was required but is not available.",
		-10829: "The application to be launched cannot run simultaneously in two different user sessions.",
	}

	def __init__(self, errornumber, apppath):
		self._number = errornumber
		self._apppath = apppath
		Exception.__init__(self, errornumber, apppath)
	
	errornumber = property(lambda self: self._number, doc="int -- Mac OS error number")
	
	apppath = property(lambda self: self._apppath, doc="str -- application path")
	
	def __int__(self):
		return self._number
	
	def __str__(self):
		return "Can't launch application at %r: %s (%i)" % (self._apppath, self._lserrors.get(self._number, 'OS error'), self._number)


def launchapp(path, newinstance=False, hide=False):
	"""Send a 'launch' event to an application. If application is not already running, it will be launched in background first.
		path : string -- full path to application, e.g. '/Applications/TextEdit.app'
		newinstance : bool -- launch a new application instance?
		hide : bool -- hide after launch?
		Result : AEAddressDesc
	"""
	if newinstance:
		desc = _launchapplication(path, _launchevent, newinstance, hide)
	else:
		try:
			# If app is already running, calling ae.launchapplication will send a 'reopen' event, so need to check for this first:
			desc = ae.psnforapplicationpath(path)
		except ae.MacOSError, err:
			if err.args[0] == -600: # Application isn't running, so launch it and send it a 'launch' event:
				sleep(1)
				desc = _launchapplication(path, _launchevent, newinstance, hide)
			else:
				raise
		else: # App is already running, so send it a 'launch' event:
			ae.newappleevent('ascr', 'noop', desc, kae.kAutoGenerateReturnID, 
					kae.kAnyTransactionID).send(kae.kAEWaitReply, kae.kAEDefaultTimeout)
	return desc

##

def processexistsforpath(path):
	"""Does a local process launched from the specified application file exist?
		Note: if path is invalid, a MacOSError is raised.
	"""
	try:
		ae.psnforapplicationpath(path)
		return True
	except ae.MacOSError, err:
		if err.args[0] == -600: 
			return False
		else:
			raise

def processexistsforpid(pid):
	"""Is there a local application process with the given unix process id?"""
	return bool(ae.isvalidpid(pid))

def processexistsforurl(url):
	"""Does an application process specified by the given eppc:// URL exist?
		Note: this will send a 'launch' Apple event to the target application.
	"""
	if ':' not in url: # workaround: process will crash if no colon in URL (OS bug)
		raise ValueError("Invalid url: %r" % url)
	return processexistsfordesc(ae.newdesc(kae.typeApplicationURL, url))

def processexistsfordesc(desc):
	"""Does an application process specified by the given AEAddressDesc exist?
		Returns false if process doesn't exist OR remote Apple events aren't allowed.
		Note: this will send a 'launch' Apple event to the target application.
	"""
	try:
		# This will usually raise error -1708 if process is running, and various errors
		# if the process doesn't exist/can't be reached. If app is running but busy,
		# AESendMessage() may return a timeout error (this should be -1712, but
		# -609 is often returned instead for some reason).
		Event(desc, 'ascrnoop').send()
	except ae.MacOSError, err:
		return err.args[0] not in [-600, -905] # -600 = no process; -905 = no network access
	return True


#######

currentapp = ae.newdesc(kae.typeProcessSerialNumber, struct.pack('II', 0, _kCurrentProcess))


def localapp(path, newinstance=False, hide=False):
	"""Make an AEAddressDesc identifying a local application. (Application will be launched if not already running.)
		path : string -- full path to application, e.g. '/Applications/TextEdit.app'
		newinstance : bool -- launch a new application instance?
		hide : bool -- hide after launch?
		Result : AEAddressDesc
	"""
	# Always create AEAddressDesc by process serial number; that way there's no confusion if multiple versions of the same app are running
	if newinstance:
		desc = _launchapplication(path, _runevent, newinstance, hide)
	else:
		try:
			desc = ae.psnforapplicationpath(path)
		except ae.MacOSError, err:
			if err.args[0] == -600: # Application isn't running, so launch it in background and send it a standard 'run' event.
				sleep(1)
				desc = _launchapplication(path, _runevent, newinstance, hide)
			else:
				raise
	return desc


def localappbypid(pid):
	"""Make an AEAddressDesc identifying a local process.
		pid : integer -- Unix process id
		Result : AEAddressDesc
	"""
	return ae.newdesc(kae.typeKernelProcessID, struct.pack('i', pid))


def remoteapp(url):
	"""Make an AEAddressDesc identifying a running application on another machine.
		url : string -- URL for remote application, e.g. 'eppc://user:password@192.168.2.1/TextEdit'
		Result : AEAddressDesc
	"""
	if ':' not in url: # workaround: process will crash if no colon in URL (OS bug)
		raise ValueError("Invalid url: %r" % url)
	return ae.newdesc(kae.typeApplicationURL, url)




"""reference -- High-level, terminology-driven wrapper for aem. 

Lots of syntactic sugar allows users to construct query-based references using familiar OO-like syntax.
"""

import struct, sys
from time import sleep

import aem, mactypes
from aem import kae
from aem.ae import newlist, newrecord, newdesc, MacOSError
from aem.aemreference import InsertionSpecifier, Test

from genericreference import GenericReference
import terminology
from terminology import kProperty, kElement
from referencerenderer import renderreference
from keywordwrapper import Keyword


######################################################################
# PRIVATE
######################################################################
# Codecs

_lowlevelcodecs = aem.Codecs()


###################################
# Application-specific data/codecs

class AppData(aem.Codecs):
	"""Provides application-specific:
		- aem.Application instance
		- name-code terminology translation tables
		- pack/unpack methods
		- help system
	"""
	
	#######
	# constants, etc.
	
	keyAECompOperator = aem.AEType(kae.keyAECompOperator)
	keyAEObject1 = aem.AEType(kae.keyAEObject1)
	keyAEObject2 = aem.AEType(kae.keyAEObject2)

	kAppscriptTypeCompDescriptorOperators = {
			kae. kAEGreaterThan: 'AS__gt__',
			kae.kAEGreaterThanEquals: 'AS__ge__',
			kae.kAEEquals: 'AS__eq__',
			kae.kAELessThan: 'AS__lt__',
			kae.kAELessThanEquals: 'AS__le__',
			kae.kAEBeginsWith: 'beginswith',
			kae.kAEEndsWith: 'endswith',
			kae.kAEContains: 'contains'
	}
	
	kClassKeyword = Keyword('class_')
	kClassType = aem.AEType('pcls')
	
	kHelpAgentBundleID = 'net.sourceforge.appscript.asdictionary'
	
	
	#######
	# initialiser
	
	def __init__(self, aemapplicationclass, constructor, identifier, terms, aemconstructoroptions={}):
		"""
			aemapplicationclass : class -- aem.Application or equivalent
			constructor : str -- indicates how to construct the aem.Application instance ('path', 'pid', 'url', 'aemapp', 'current')
			identifier : any -- value identifying the target application (its type is dependent on constructor parameter)
			terms : bool | module | tuple -- if True, retrieve terminology from target application dynamically; 
					if false, use only default terminology; if static glue module or tuple, use terminology from that
			aemconstructoroptions -- any additional keyword arguments to pass to aemapplicationclass constructor
					(e.g. newinstance, hide)
		"""
		# initialise codecs
		aem.Codecs.__init__(self)
		self.decoders.update({
				kae.typeType: self.unpackkeyword,
				kae.typeEnumerated: self.unpackkeyword,
				kae.typeProperty: self.unpackkeyword,
				kae.typeObjectSpecifier: self.unpackreference,
				kae.typeInsertionLoc: self.unpackreference,
				# AEAddressDesc types
				kae.typeApplicationBundleID: self.unpackapplicationbyid,
				kae.typeApplicationURL: self.unpackapplicationbyurl,
				kae.typeApplSignature: self.unpackapplicationbysignature,
				kae.typeKernelProcessID: self.unpackapplicationbypid,
				kae.typeMachPort: self.unpackapplicationbydesc,
				kae.typeProcessSerialNumber: self.unpackapplicationbydesc,
		})
		# store parameters for later use
		self._aemapplicationclass = aemapplicationclass
		self.constructor, self.identifier = constructor, identifier
		self.aemconstructoroptions = aemconstructoroptions
		self._relaunchmode = 'limited'
		self._terms = terms
		self._helpagent = None
		self._isconnected = False
	
	
	#######
	
	# Terminology-aware pack/unpack functions.
	# These replace the default aem pack/unpack functions, which don't understand appscript Keyword and Reference objects.
	
	def packdict(self, val):
		# Pack dictionary whose keys are strings (e.g. 'foo'), Keywords (e.g. k.name) or AETypes (e.g. AEType('pnam').
		record = newrecord()
		if self.kClassKeyword in val or self.kClassType in val:
			# if hash contains a 'class' property containing a class name, coerce the AEDesc to that class
			newval = val.copy()
			if self.kClassKeyword in newval:
				value = newval.pop(self.kClassKeyword)
			else:
				value = newval.pop(self.kClassType)
			if isinstance(value, Keyword): # get the corresponding AEType (assuming there is one)
				value = self.typebyname().get(value.name, value)
			if isinstance(value, aem.AEType): # coerce the record to the desired type
				record = record.coerce(value.code)
				val = newval
		usrf = None
		for key, value in val.items():
			if isinstance(key, Keyword):
				try:
					keyCode = self.typebyname()[key.AS_name].code
				except KeyError:
					raise ValueError("Unknown Keyword: k.%s" % key.AS_name)
				record.setparam(keyCode, self.pack(value))
			elif isinstance(key, aem.AETypeBase): # AEType/AEProp (AEType is normally used in practice)
				record.setparam(key.code, self.pack(value))
			else: # user-defined key (normally a string)
				if not usrf:
					usrf = newlist()
				usrf.setitem(0, self.pack(key))
				usrf.setitem(0, self.pack(value))
		if usrf:
			record.setparam('usrf', usrf)
		return record
	
	
	def unpackaerecord(self, desc):
		# Unpack typeAERecord,  converting record keys to Keyword objects (not AETypes) where possible.
		dct = {}
		for i in range(desc.count()):
			key, value = desc.getitem(i + 1, kae.typeWildCard)
			if key == 'usrf':
				lst = self.unpack(value)
				for i in range(0, len(lst), 2):
					dct[lst[i]] = lst[i+1]
			elif key in self.typebycode():
				dct[self.typebycode()[key]] = self.unpack(value)
			else:
				dct[aem.AEType(key)] = self.unpack(value)
		return dct

	##
	
	def unpackkeyword(self, desc):
		# Unpack typeType, typeEnum, typeProperty; replaces default aem decoders to convert types, enums, etc.
		# to Keyword objects instead of AETypes, AEEnums, etc.
		aemValue = _lowlevelcodecs.unpack(desc)
		return self.typebycode().get(aemValue.code, aemValue)
	
	
	def unpackreference(self, desc):
		return Reference(self, _lowlevelcodecs.unpack(desc))
	
	
	def unpackcompdescriptor(self, desc):
		# need to do some typechecking when unpacking 'contains' comparisons, so have to override the low-level unpacker
		rec = self.unpack(desc.coerce(kae.typeAERecord))
		operator = self.kAppscriptTypeCompDescriptorOperators[rec[self.keyAECompOperator].code]
		op1 = rec[self.keyAEObject1]
		op2 = rec[self.keyAEObject2]
		if operator == 'contains':
			if isinstance(op1, Reference) and op1.AS_aemreference.AEM_root() == aem.its:
				return op1.contains(op2)
			elif isinstance(op2, Reference) and op2.AS_aemreference.AEM_root() == aem.its:
				return op2.isin(op1)
			else:
				return _lowlevelcodecs.unpack(desc)
		else:
			return getattr(op1, operator)(op2)
	
	##
	
	def unpackapplicationbyid(self, desc):
		return app(id=desc.data)
	
	def unpackapplicationbyurl(self, desc):
		if desc.data.startswith('file'): # workaround for converting AEAddressDescs containing file:// URLs to application paths, since AEAddressDescs containing file URLs don't seem to work correctly
			return app(mactypes.File.makewithurl(desc.data).path)
		else: # presumably contains an eppc:// URL
			return app(url=desc.data)
	
	def unpackapplicationbysignature(self, desc):
		return app(creator=struct.pack('>I', struct.unpack('I', desc.data)[0]))
	
	def unpackapplicationbypid(self, desc):
		return app(pid=struct.unpack('I', desc.data)[0])
	
	def unpackapplicationbydesc(self, desc):
		return app(aemapp=aem.Application(desc=desc))

	#######
	
	isconnected = property(lambda self: self._isconnected)
	
	def connect(self):
		"""Initialises application target and terminology lookup tables.
		
		Called automatically the first time clients retrieve target, typebycode, typebyname,
		referencebycode, referencebyname; clients should not need to call it themselves.
		"""
		# initialise target (by default an aem.Application instance)
		if self.constructor == 'aemapp':
			t = self._target = self.identifier
		elif self.constructor == 'current':
			t = self._target = self._aemapplicationclass()
		else:
			kargs = {self.constructor: self.identifier}
			kargs.update(self.aemconstructoroptions)
			t = self._target = self._aemapplicationclass(**kargs)
		# initialise translation tables
		if self._terms == True: # obtain terminology from application
			self._terms = terminology.tablesforapp(t)
		elif self._terms == False: # use built-in terminology only (e.g. use this when running AppleScript applets)
			self._terms = terminology.defaulttables
		elif not isinstance(self._terms, tuple): # use user-supplied terminology module
			self._terms = terminology.tablesformodule(self._terms)
		d1, d2, d3, d4 = self._typebycode, self._typebyname, \
				self._referencebycode, self._referencebyname = self._terms
		self.target = lambda: t
		self.typebycode = lambda: d1
		self.typebyname = lambda: d2
		self.referencebycode = lambda: d3
		self.referencebyname = lambda: d4
		self._isconnected = True
	
	def target(self):
		self.connect()
		return self._target
	
	def typebycode(self):
		self.connect()
		return self._typebycode
	
	def typebyname(self):
		self.connect()
		return self._typebyname
	
	def referencebycode(self):
		self.connect()
		return self._referencebycode
	
	def referencebyname(self):
		self.connect()
		return self._referencebyname
	
	##
	
	def pack(self, data):
		if isinstance(data, GenericReference):
			data = data.AS_resolve(Reference, self)
		if isinstance(data, Reference):
			data = data.AS_aemreference
		elif isinstance(data, Keyword):
			try:
				data = self.typebyname()[data.AS_name]
			except KeyError:
				raise ValueError("Unknown Keyword: k.%s" % data.AS_name)
		return aem.Codecs.pack(self, data)
	
	# Relaunch mode
	
	def _setrelaunchmode(self, mode):
		if mode not in ['never', 'limited', 'always']:
			raise ValueError('Unknown relaunch mode: %r' % mode)
		self._relaunchmode = mode
	relaunchmode = property(lambda self: self._relaunchmode, _setrelaunchmode)
	
	# Help system
	
	def _write(self, s):
		if isinstance(s, unicode):
			s= s.encode('utf8')
		print >> sys.stderr, s
	
	def _inithelpagent(self):
		try:
			apppath = aem.findapp.byid(self.kHelpAgentBundleID)
			self._helpagent = aem.Application(apppath, hide=True)
			return True
		except aem.findapp.ApplicationNotFoundError:
			self._write("No help available: ASDictionary application not found.")
		except aem.CantLaunchApplicationError:
			self._write("No help available: can't launch ASDictionary application.")
		return False
	
	def _displayhelp(self, flags, ref):
		if isinstance(ref, Command):
			commandname = ref.AS_name
			ref = ref.AS_aemreference
		else:
			commandname = ''
		try:
			self._write(self._helpagent.event('AppSHelp', {
					'Cons': self.constructor,
					'Iden': self.identifier,
					'Styl': 'py-appscript',
					'Flag': flags,
					'aRef': self.pack(ref),
					'CNam': commandname
					}).send())
			return None
		except aem.EventError, e:
			return e
	
	def help(self, flags, ref):
		try:
			if not self._helpagent: # initialise help system upon first use
				if not self._inithelpagent():
					return ref # if ASDictionary is unavailable then do nothing
			e = self._displayhelp(flags, ref)
			if e and e.errornumber in [-600, -609]: # ASDictionary is no longer running, so reconnect
				if not self._inithelpagent():
					return ref # if ASDictionary is unavailable then do nothing
				e = self._displayhelp(flags, ref)
			if e:
				self._write("No help available: ASDictionary raised an error: %s" % e)
		except Exception, err:
			self._write("No help available: unknown error: %s" % err)
			from traceback import print_exc
			print_exc()
		return ref


######################################################################
# Considering/ignoring constants

def _packuint32(n): # used to pack csig attributes
	return newdesc(kae.typeUInt32, struct.pack('I', n))

# 'csig' attribute flags (see ASRegistry.h; note: there's no option for 'numeric strings' in 10.4)

_ignoreenums = [
	(Keyword('case'), kae.kAECaseConsiderMask, kae.kAECaseIgnoreMask),
	(Keyword('diacriticals'), kae.kAEDiacriticConsiderMask, kae.kAEDiacriticIgnoreMask),
	(Keyword('whitespace'), kae.kAEWhiteSpaceConsiderMask, kae.kAEWhiteSpaceIgnoreMask),
	(Keyword('hyphens'), kae.kAEHyphensConsiderMask, kae.kAEHyphensIgnoreMask),
	(Keyword('expansion'), kae.kAEExpansionConsiderMask, kae.kAEExpansionIgnoreMask),
	(Keyword('punctuation'), kae.kAEPunctuationConsiderMask, kae.kAEPunctuationIgnoreMask),
	]

# default cons, csig attributes

_defaultconsiderations =  _lowlevelcodecs.pack([aem.AEEnum(kae.kAECase)])
_defaultconsidsandignores = _packuint32(kae.kAECaseIgnoreMask)


######################################################################
# Base class for references and commands

class _Base(object):
	# Base class for Command and Reference objects.
	def __init__(self, appdata):
		self.AS_appdata = appdata
		
	# Help system
	
	def help(self, flags='-t'): # add a help() method to all concrete app, reference and command objects
		"""Print help. Use help('-h') for more info."""
		return self.AS_appdata.help(flags, self)


######################################################################
# PUBLIC
######################################################################
# The Reference and Command classes are used to construct references and commands applying to those references

class Command(_Base):
	
	def __init__(self, parentref, name, code, arginfo):
		_Base.__init__(self, parentref.AS_appdata)
		self.AS_aemreference = parentref.AS_aemreference
		self.AS_name = name
		self._parentref = parentref
		self._code = code
		self._labelledargterms = arginfo
	
	def __repr__(self):
		return '%r.%s' % (self._parentref, self.AS_name)
	
	def __call__(self, *args, **kargs):
		keywordargs = kargs.copy()
		if len(args) > 1:
			raise TypeError("Command received more than one direct parameter %r." % (args,))
		# get user-specified timeout, if any
		timeout = int(keywordargs.pop('timeout', 60)) # appscript's default is 60 sec
		if timeout <= 0:
			timeout = kae.kNoTimeOut
		else:
			timeout *= 60 # convert to ticks
		# ignore application's reply?
		sendflags = keywordargs.pop('waitreply', True) and kae.kAEWaitReply or kae.kAENoReply
		atts, params = {'subj':None}, {}
		# add considering/ignoring attributes (note: most apps currently ignore these)
		ignoreoptions = keywordargs.pop('ignore', None)
		if ignoreoptions is None:
			atts['cons'] = _defaultconsiderations # 'csig' obsoletes 'cons', but latter is retained for compatibility
			atts['csig'] = _defaultconsidsandignores
		else:
			atts['cons'] = ignoreoptions
			csig = 0
			for option, considermask, ignoremask in _ignoreenums:
				csig += option in ignoreoptions and ignoremask or considermask
			atts['csig'] = _packuint32(csig)
		# optionally have application supply return value as specified type
		if 'resulttype' in keywordargs:
			params['rtyp'] = keywordargs.pop('resulttype')
		# add direct parameter, if any
		if args:
			params['----'] = args[0]
		# extract Apple event's labelled parameters, if any
		try:
			for name, value in keywordargs.items():
				params[self._labelledargterms[name]] = value
		except KeyError:
			raise TypeError('Unknown keyword argument %r.' % name)
		# apply special cases for certain commands (make, set, any command that takes target object specifier as its direct parameter); appscript provides these as a convenience to users, making its syntax more concise, OO-like and nicer to use
		if self.AS_aemreference is not aem.app:
			if self._code == 'coresetd':
				# Special case: if ref.set(...) contains no 'to' argument, use direct argument for 'to' parameter and target reference for direct parameter
				if '----' in params and 'data' not in params:
					params['data'] = params['----']
					params['----'] = self.AS_aemreference
				elif '----' not in params:
					params['----'] = self.AS_aemreference
				else:
					atts['subj'] = self.AS_aemreference
			elif self._code == 'corecrel':
				# this next bit is a bit tricky: 
				# - While it should be possible to pack the target reference as a subject attribute, when the target is of typeInsertionLoc, CocoaScripting stupidly tries to coerce it to typeObjectSpecifier, which causes a coercion error.
				# - While it should be possible to pack the target reference as the 'at' parameter, some less-well-designed applications won't accept this and require it to be supplied as a subject attribute (i.e. how AppleScript supplies it).
				# One option is to follow the AppleScript approach and force users to always supply subject attributes as target references and 'at' parameters as 'at' parameters, but the syntax for the latter is clumsy and not backwards-compatible with a lot of existing appscript code (since earlier versions allowed the 'at' parameter to be given as the target reference). So for now we split the difference when deciding what to do with a target reference: if it's an insertion location then pack it as the 'at' parameter (where possible), otherwise pack it as the subject attribute (and if the application doesn't like that then it's up to the client to pack it as an 'at' parameter themselves).
				#
				# if ref.make(...) contains no 'at' argument and target is an insertion reference, use target reference for 'at' parameter...
				if isinstance(self.AS_aemreference, InsertionSpecifier) and 'insh' not in params:
					params['insh'] = self.AS_aemreference
				else: # ...otherwise pack the target reference as the subject attribute
					atts['subj'] = self.AS_aemreference
			elif '----' in params:
				# if user has already supplied a direct parameter, pack that reference as the subject attribute
				atts['subj'] = self.AS_aemreference
			else:
				# pack that reference as the direct parameter
				params['----'] = self.AS_aemreference
		# build and send the Apple event, returning its result, if any
		try:
			return self.AS_appdata.target().event(self._code, params, atts, codecs=self.AS_appdata).send(timeout, sendflags)
		except aem.EventError, e:
			if e.errornumber == -1708 and self._code == 'ascrnoop':
				return # 'launch' events always return 'not handled' errors; just ignore these
			elif e.errornumber in [-600, -609] and self.AS_appdata.constructor == 'path':
				#
				# Event was sent to a local app for which we no longer have a valid address
				# (i.e. the application has quit since this aem.Application object was made).
				#
				# - If application is running under a new process id, we just update the 
				#   aem.Application object and resend the event.
				#
				# - If application isn't running, then we see if the event being sent is one of 
				#   those allowed to relaunch the application (i.e. 'run' or 'launch'). If it is, the
				#   application is relaunched, the process id updated and the event resent;
				#   if not, the error is rethrown.
				#
				if not self.AS_appdata.target().processexistsforpath(self.AS_appdata.identifier):
					if self.AS_appdata.relaunchmode == 'never':
						raise CommandError(self, (args, kargs), e, self.AS_appdata)
					elif self.AS_appdata.relaunchmode == 'limited':
						if self._code == 'ascrnoop':
							aem.Application.launch(self.AS_appdata.identifier) # relaunch app in background
						elif self._code != 'aevtoapp': # only 'launch' and 'run' are allowed to restart a local application that's been quit
							raise CommandError(self, (args, kargs), e, self.AS_appdata)
					# else always relaunch
				# update AEMApplication object's AEAddressDesc
				self.AS_appdata.target().reconnect()
				# re-send command
				try:
					return self.AS_appdata.target().event(self._code, params, atts, 
							codecs=self.AS_appdata).send(timeout, sendflags)
				except aem.EventError, e:
					if e.errornumber == -1708 and self._code == 'ascrnoop':
						return
					raise CommandError(self, (args, kargs), e, self.AS_appdata)
			raise CommandError(self, (args, kargs), e, self.AS_appdata)
	
	def AS_formatcommand(self, args):
		return '%r(%s)' % (self, ', '.join(['%r' % (v,) for v in args[0]] + ['%s=%r' % (k, v) for (k, v) in args[1].items()]))
		

######################################################################

class Reference(_Base):
	# A general-purpose class used to construct all real appscript references. 
	# It's a simple wrapper around an aem reference that provides syntactic sugar 
	# and terminology->AE code conversion. Calling a reference-building method 
	# returns a new Reference object containing the new reference, except where 
	# it would create a structurally invalid reference (e.g. ref.items[1]['foo']), in which 
	# case the aem reference will raise an AttributeError.
	
	def __init__(self, appdata, aemreference):
		_Base.__init__(self, appdata)
		self.AS_aemreference = aemreference # an aem app-/con-/its-based reference
	
	def _resolverangeboundary(self, selector, valueifnone): # used by __getitem__() below
		if selector is None: # e.g. documents[2:]
			selector = valueifnone 
		if isinstance(selector, GenericReference):
			return selector.AS_resolve(Reference, self.AS_appdata).AS_aemreference
		elif isinstance(selector, Reference):
			return selector.AS_aemreference
		else:
			return selector
	
	# Full references are hashable and comparable for equality. (Generic references aren't, 
	# however, as __eq__() is overridden for other purposes, but the user shouldn't be 
	# troubled by this given how generic refs are normally used.)
	
	def __eq__(self, val):
		return self.__class__ == val.__class__ and \
				self.AS_appdata.target() == val.AS_appdata.target() and \
				self.AS_aemreference == val.AS_aemreference
	
	def __ne__(self, val):
		return not self == val
	
	def __hash__(self):
		val = hash((self.AS_aemreference, self.AS_appdata.target()))
		self.__hash__ = lambda: val
		return val
	
	def __iter__(self): # dummy-proof
		raise RuntimeError("Can't iterate an application reference; use ref.get() to return a list of references first.")
	
	def __repr__(self): # references display as themselves
		val = renderreference(self.AS_appdata, self.AS_aemreference)
		self.__repr__ = lambda: val
		return val
		
	# miscellaneous
	
	def isrunning(self):
		constructor = self.AS_appdata.constructor
		identifier = self.AS_appdata.identifier
		if constructor == 'path':
			return aem.Application.processexistsforpath(identifier)
		elif constructor == 'pid':
			return aem.Application.processexistsforpid(identifier)
		elif constructor == 'url':
			return aem.Application.processexistsforurl(identifier)
		elif constructor == 'aemapp':
			return aem.Application.processexistsforpath(identifier.addressdesc)
		else: # constructor == 'current'
			return True
	
	def _setrelaunchmode(self, mode): # mode = 'never', 'limited' or 'always'
		self.AS_appdata.relaunchmode = mode
	relaunchmode = property(lambda self: self.AS_appdata.relaunchmode, _setrelaunchmode)
	
	# Public properties and methods; these are called by end-user and other clients (e.g. generic references)
	
	def __getattr__(self, name):
		try:
			selectortype, code = self.AS_appdata.referencebyname()[name]
		except KeyError:
			raise AttributeError("Unknown property, element or command: %r" % name)
		if selectortype == kProperty:
			return Reference(self.AS_appdata, self.AS_aemreference.property(code))
		elif selectortype == kElement:
			return Reference(self.AS_appdata, self.AS_aemreference.elements(code))
		else: # kCommand (note: 'code' variable here actually contains a (code, args) struct)
			return Command(self, name, code[0], code[1])
	
	def __getitem__(self, selector):
		if isinstance(selector, basestring): # by-name
			return Reference(self.AS_appdata, self.AS_aemreference.byname(selector))
		elif isinstance(selector, (GenericReference, Reference, Test)): # by-test
			if isinstance(selector, GenericReference):
				testclause = selector.AS_resolve(Reference, self.AS_appdata)
				try:
					testclause = testclause.AS_aemreference
				except AttributeError:
					raise ValueError('Not a valid its-based test: %r' % selector)
			elif isinstance(selector, Reference):
				testclause = selector.AS_aemreference
			else:
				testclause = selector
			if not isinstance(testclause, Test):
				raise TypeError('Not an its-based test: %r' % selector)
			return Reference(self.AS_appdata, self.AS_aemreference.byfilter(testclause))
		elif isinstance(selector, slice): # by-range
			return Reference(self.AS_appdata, self.AS_aemreference.byrange(
					self._resolverangeboundary(selector.start, 1),
					self._resolverangeboundary(selector.stop, -1)))
		else: # by-index
			return Reference(self.AS_appdata, self.AS_aemreference.byindex(selector))
	
	first = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.first))
	middle = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.middle))
	last = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.last))
	any = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.any))
	beginning = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.beginning))
	end = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.end))
	before = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.before))
	after = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.after))
	
	def previous(self, klass):
		try:
			aemtype = self.AS_appdata.typebyname()[klass.AS_name]
		except AttributeError: # can't get klass.AS_name
			raise TypeError("Not a keyword: %r" % name)
		except KeyError: # can't get typebyname[<name>]
			raise ValueError("Unknown class: %r" % name)
		return Reference(self.AS_appdata, self.AS_aemreference.previous(aemtype.code))
	
	def next(self, klass):
		try:
			aemtype = self.AS_appdata.typebyname()[klass.AS_name]
		except AttributeError: # can't get klass.AS_name
			raise TypeError("Not a keyword: %r" % name)
		except KeyError: # can't get typebyname[<name>]
			raise ValueError("Unknown class: %r" % name)
		return Reference(self.AS_appdata, self.AS_aemreference.next(aemtype.code))
	
	def ID(self, id):
		return Reference(self.AS_appdata, self.AS_aemreference.byid(id))
	
	def __call__(self, *args, **kargs):
		return self.get(*args, **kargs)
	
	# Following methods will be called by its-based generic references when resolving themselves into real references; end-users and other clients shouldn't call them directly.
	
	def AS__gt__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.gt(operand))
	
	def AS__ge__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.ge(operand))
	
	def AS__eq__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.eq(operand))
	
	def AS__ne__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.ne(operand))
	
	def AS__lt__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.lt(operand))
	
	def AS__le__(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.le(operand))
	
	def beginswith(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.beginswith(operand))
	
	def endswith(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.endswith(operand))
	
	def contains(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.contains(operand))
	
	def isin(self, operand):
		return Reference(self.AS_appdata, self.AS_aemreference.isin(operand))
	
	def doesnotbeginwith(self, operand):
		return self.beginswith(operand).NOT
	
	def doesnotendwith(self, operand):
		return self.endswith(operand).NOT
	
	def doesnotcontain(self, operand):
		return self.contains(operand).NOT
	
	def isnotin(self, operand):
		return self.isin(operand).NOT
	
	def AND(self, *operands):
		return Reference(self.AS_appdata, self.AS_aemreference.AND(*operands))
		
	def OR(self, *operands):
		return Reference(self.AS_appdata, self.AS_aemreference.OR(*operands))
	
	NOT = property(lambda self: Reference(self.AS_appdata, self.AS_aemreference.NOT))


######################################################################
# The Application class is not directly instantiated by the user; instead, they call 
# the GenericApp instance and this returns the real Application instance. This allows 
# users to write generic app-based references, e.g. app.documents.end, as well as 
# real ones, e.g. app('TextEdit').documents.end

class Application(Reference):
	"""Creates objects for communicating with scriptable applications."""
	
	# overridable hook; appscript.Application subclasses can modify creating
	# and/or sending Apple events by using custom aem.Application and
	# aem.Event classes # Note: subclassing this class is now a bit trickier 
	# due to introduction of generic 'app'; clients need to import this class 
	# directly, subclass it, and then create their own GenericApp instance to 
	# use in place of the standard version.
	_Application = aem.Application
	
	def __init__(self, name=None, id=None, creator=None, pid=None, url=None, aemapp=None, 
			terms=True, newinstance=False, hide=False):
		"""
			app(name=None, id=None, creator=None, pid=None, url=None, terms=True)
				name : str -- name or path of application, e.g. 'TextEdit', 'TextEdit.app', '/Applications/Textedit.app'
				id : str -- bundle id of application, e.g. 'com.apple.textedit'
				creator : str -- 4-character creator type of application, e.g. 'ttxt'
				pid : int -- Unix process id, e.g. 955
				url : str -- eppc:// URL, e.g. eppc://G4.local/TextEdit'
				aemapp : aem.Application
				terms : module | bool -- if a module, get terminology from it; if True, get terminology 
						from target application; if False, use built-in terminology only
				newinstance : bool -- launch a new application instance?
				hide : bool -- hide after launch?
			
			Notes: 
			
			- The newinstance and hide options only apply when specifying application by name, id, or creator.
		"""
		if len([i for i in [name, id, creator, pid, url, aemapp] if i]) > 1:
			raise TypeError('app() received more than one of the following arguments: name, id, creator, pid, url, aemapp')
		if name:
			constructor, identifier = 'path', aem.findapp.byname(name)
		elif id:
			constructor, identifier = 'path',  aem.findapp.byid(id)
		elif creator:
			constructor, identifier = 'path',  aem.findapp.bycreator(creator)
		elif pid:
			constructor, identifier = 'pid', pid
		elif url:
			constructor, identifier = 'url', url
		elif aemapp:
			constructor, identifier = 'aemapp', aemapp
		else:
			constructor, identifier = 'current', None
		# Defer initialisation of AppData until it's needed. This allows user to call launch() on a non-running application without the application being launched by aem.Application, which automatically launches local applications in order to construct an AEAddressDesc of typeProcessSerialNumber.
		# launch()'s usefulness is somewhat limited, since constructing a real app-based reference will also launch the application normally in order to get its terminology. So to actually launch an application, you have to use launch() before constructing any real references to its objects; i.e.:
		#     te = app('TextEdit'); te.launch(); d = app.documents
		# will launch TE without it creating any new documents (i.e. app receives 'ascrnoop' as its first event), but:
		#     te = app('TextEdit'); d = app.documents; te.launch()
		# will launch TE normally (i.e. app receives 'aevtoapp' as its first event), causing it to open a new, empty window.
		Reference.__init__(self, 
				AppData(self._Application, constructor, identifier, terms, {'newinstance': newinstance, 'hide': hide}), 
				aem.app)
	
	def AS_newreference(self, ref):
		"""Create a new appscript reference from an aem reference."""
		if isinstance(ref, GenericReference):
			return ref.AS_resolve(Reference, self.AS_appdata)
		elif isinstance(ref, aem.Query):
			return Reference(self.AS_appdata, ref)
		elif ref is None:
			return Reference(self.AS_appdata, aem.app)
		else:
			return Reference(self.AS_appdata, aem.customroot(ref))

	def begintransaction(self, session=None):
		self.AS_appdata.target().begintransaction(session)
	
	def aborttransaction(self):
		self.AS_appdata.target().aborttransaction()
	
	def endtransaction(self):
		self.AS_appdata.target().endtransaction()
	
	def launch(self):
		"""Launch a non-running application in the background and send it a 'launch' event. 
			Note: this will only launch non-running apps that are specified by name/path/
			bundle id/creator type. Apps specified by other means will be still sent a 
			launch event if already running, but an error will occur if they're not.
		"""
		if not self.isrunning() and self.AS_appdata.constructor == 'path' \
				and self.AS_appdata.relaunchmode != 'never':
			aem.Application.launch(self.AS_appdata.identifier)
			self.AS_appdata.target().reconnect() # make sure aem.Application object's AEAddressDesc is up to date
		else: # send launch event to app (will error if not already running)
			Command(self, 'launch', 'ascrnoop', {})()


#######

class GenericApp(GenericReference):
	def __init__(self, appclass):
		self._appclass = appclass
		GenericReference.__init__(self, ['app'])
		
	def __call__(self, *args, **kargs):
		return self._appclass(*args, **kargs)


app = GenericApp(Application) # app-based references are generic references unless you specify an application by calling app, e.g. app.home is generic, app('Finder').home is real


######################################################################
# The CommandError class is exposed for use in try...except... blocks

class CommandError(Exception):
	"""An error raised when sending a command (e.g. aem.EventError). Contains
		the original exception object and the command on which the error was raised.
	"""
	
	def __init__(self, command, parameters, realerror, codecs):
		"""
			command : Command -- command reference
			parameters : tuple -- two-item tuple containing tuple of positional args and dict of keyword args
			realerror : Exception -- the original error raised
			codecs : AppData -- used to unpack error data as needed
		"""
		self.command, self. parameters, self.realerror = command, parameters, realerror
		self._codecs = codecs
		Exception.__init__(self, command, parameters, realerror)
		
	def __int__(self):
		if isinstance(self.realerror, aem.EventError):
			return int(self.realerror)
		elif isinstance(self.realerror, MacOSError):
			return self.realerror[0]
		else:
			return -2700
	
	def __repr__(self):
		return 'appscript.CommandError(%r, %r, %r)' % (self.command, self. parameters, self.realerror)
	
	def __str__(self):
		if isinstance(self.realerror, aem.EventError):
			err = "Command failed:\n\t\tOSERROR: %i" % self.errornumber
			msg = self.errormessage
			if msg:
				err += "\n\t\tMESSAGE: %s" % msg
			for label, key in [
					["OFFENDING OBJECT", kae.kOSAErrorOffendingObject], 
					["EXPECTED TYPE", kae.kOSAErrorExpectedType],
					["PARTIAL RESULT", kae.kOSAErrorPartialResult]]:
				if key in self.realerror.raw:
					err += "\n\t\t%s: %r" % (label, self._codecs.unpack(self.realerror.raw[key]))
		else:
			err = self.realerror
		return "%s\n\t\tCOMMAND: %s" % (err, self.command.AS_formatcommand(self.parameters))
	
	# basic error info (an error number is always given by AEM/application;
	# message is either supplied by application or generated by aem.EventError)
	
	errornumber = property(lambda self: int(self), doc="int -- Mac OS error number")
	
	def _errormessage(self):
		if isinstance(self.realerror, aem.EventError):
			return self.realerror.errormessage
		else:
			return ''
	errormessage = property(_errormessage, doc="str -- application-supplied/generic error description")

	# extended error info (some apps may return some or all additional error info, though most don't)
	
	def _errorinfo(self, key):
		if isinstance(self.realerror, aem.EventError):
			desc = self.realerror.raw.get(key)
			if desc:
				return self._codecs.unpack(desc)
		return None
	
	offendingobject = property(lambda self: self._errorinfo(kae.kOSAErrorOffendingObject),
			doc="anything | None -- object that caused the error, if given by application")
	expectedtype = property(lambda self: self._errorinfo(kae.kOSAErrorExpectedType),
			doc="anything | None -- object that caused a coercion error, if given by application")
	partialresult = property(lambda self: self._errorinfo(kae.kOSAErrorPartialResult),
			doc="anything | None -- part of return value constructed before error occurred, if given by application")





# kae.py
#
# Generated on Sun Dec 30 17:39:53 +0000 2007


# AEDataModel.h

typeBoolean = 'bool'
typeChar = 'TEXT'

typeStyledUnicodeText = 'sutx'
typeEncodedString = 'encs'
typeUnicodeText = 'utxt'
typeCString = 'cstr'
typePString = 'pstr'

typeUTF16ExternalRepresentation = 'ut16'
typeUTF8Text = 'utf8'

typeSInt16 = 'shor'
typeUInt16 = 'ushr'
typeSInt32 = 'long'
typeUInt32 = 'magn'
typeSInt64 = 'comp'
typeUInt64 = 'ucom'
typeIEEE32BitFloatingPoint = 'sing'
typeIEEE64BitFloatingPoint = 'doub'
type128BitFloatingPoint = 'ldbl'
typeDecimalStruct = 'decm'

typeSMInt = typeSInt16
typeShortInteger = typeSInt16
typeInteger = typeSInt32
typeLongInteger = typeSInt32
typeMagnitude = typeUInt32
typeComp = typeSInt64
typeSMFloat = typeIEEE32BitFloatingPoint
typeShortFloat = typeIEEE32BitFloatingPoint
typeFloat = typeIEEE64BitFloatingPoint
typeLongFloat = typeIEEE64BitFloatingPoint
typeExtended = 'exte'

typeAEList = 'list'
typeAERecord = 'reco'
typeAppleEvent = 'aevt'
typeEventRecord = 'evrc'
typeTrue = 'true'
typeFalse = 'fals'
typeAlias = 'alis'
typeEnumerated = 'enum'
typeType = 'type'
typeAppParameters = 'appa'
typeProperty = 'prop'
typeFSRef = 'fsrf'
typeFileURL = 'furl'
typeKeyword = 'keyw'
typeSectionH = 'sect'
typeWildCard = '****'
typeApplSignature = 'sign'
typeQDRectangle = 'qdrt'
typeFixed = 'fixd'
typeProcessSerialNumber = 'psn '
typeApplicationURL = 'aprl'
typeNull = 'null'

typeFSS = 'fss '

typeCFAttributedStringRef = 'cfas'
typeCFMutableAttributedStringRef = 'cfaa'
typeCFStringRef = 'cfst'
typeCFMutableStringRef = 'cfms'
typeCFArrayRef = 'cfar'
typeCFMutableArrayRef = 'cfma'
typeCFDictionaryRef = 'cfdc'
typeCFMutableDictionaryRef = 'cfmd'
typeCFNumberRef = 'cfnb'
typeCFBooleanRef = 'cftf'
typeCFTypeRef = 'cfty'

typeKernelProcessID = 'kpid'
typeMachPort = 'port'

typeApplicationBundleID = 'bund'

keyTransactionIDAttr = 'tran'
keyReturnIDAttr = 'rtid'
keyEventClassAttr = 'evcl'
keyEventIDAttr = 'evid'
keyAddressAttr = 'addr'
keyOptionalKeywordAttr = 'optk'
keyTimeoutAttr = 'timo'
keyInteractLevelAttr = 'inte'
keyEventSourceAttr = 'esrc'
keyMissedKeywordAttr = 'miss'
keyOriginalAddressAttr = 'from'
keyAcceptTimeoutAttr = 'actm'
keyReplyRequestedAttr = 'repq'

kAEDebugPOSTHeader = (1 << 0)
kAEDebugReplyHeader = (1 << 1)
kAEDebugXMLRequest = (1 << 2)
kAEDebugXMLResponse = (1 << 3)
kAEDebugXMLDebugAll = 0xFFFFFFFF

kSOAP1999Schema = 'ss99'
kSOAP2001Schema = 'ss01'

keyUserNameAttr = 'unam'
keyUserPasswordAttr = 'pass'
keyDisableAuthenticationAttr = 'auth'
keyXMLDebuggingAttr = 'xdbg'
kAERPCClass = 'rpc '
kAEXMLRPCScheme = 'RPC2'
kAESOAPScheme = 'SOAP'
kAESharedScriptHandler = 'wscp'
keyRPCMethodName = 'meth'
keyRPCMethodParam = 'parm'
keyRPCMethodParamOrder = '/ord'
keyAEPOSTHeaderData = 'phed'
keyAEReplyHeaderData = 'rhed'
keyAEXMLRequestData = 'xreq'
keyAEXMLReplyData = 'xrep'
keyAdditionalHTTPHeaders = 'ahed'
keySOAPAction = 'sact'
keySOAPMethodNameSpace = 'mspc'
keySOAPMethodNameSpaceURI = 'mspu'
keySOAPSchemaVersion = 'ssch'

keySOAPStructureMetaData = '/smd'
keySOAPSMDNamespace = 'ssns'
keySOAPSMDNamespaceURI = 'ssnu'
keySOAPSMDType = 'sstp'

kAEUseHTTPProxyAttr = 'xupr'
kAEHTTPProxyPortAttr = 'xhtp'
kAEHTTPProxyHostAttr = 'xhth'

kAESocks4Protocol = 4
kAESocks5Protocol = 5

kAEUseSocksAttr = 'xscs'
kAESocksProxyAttr = 'xsok'
kAESocksHostAttr = 'xshs'
kAESocksPortAttr = 'xshp'
kAESocksUserAttr = 'xshu'
kAESocksPasswordAttr = 'xshw'

kAEDescListFactorNone = 0
kAEDescListFactorType = 4
kAEDescListFactorTypeAndSize = 8

kAutoGenerateReturnID = -1
kAnyTransactionID = 0

kAEDataArray = 0
kAEPackedArray = 1
kAEDescArray = 3
kAEKeyDescArray = 4

kAEHandleArray = 2

kAENormalPriority = 0x00000000
kAEHighPriority = 0x00000001

kAENoReply = 0x00000001
kAEQueueReply = 0x00000002
kAEWaitReply = 0x00000003
kAEDontReconnect = 0x00000080
kAEWantReceipt = 0x00000200
kAENeverInteract = 0x00000010
kAECanInteract = 0x00000020
kAEAlwaysInteract = 0x00000030
kAECanSwitchLayer = 0x00000040
kAEDontRecord = 0x00001000
kAEDontExecute = 0x00002000
kAEProcessNonReplyEvents = 0x00008000

kAEDefaultTimeout = -1
kNoTimeOut = -2



# AEHelpers.h

aeBuildSyntaxNoErr = 0
aeBuildSyntaxBadToken = 1
aeBuildSyntaxBadEOF = 2
aeBuildSyntaxNoEOF = 3
aeBuildSyntaxBadNegative = 4
aeBuildSyntaxMissingQuote = 5
aeBuildSyntaxBadHex = 6
aeBuildSyntaxOddHex = 7
aeBuildSyntaxNoCloseHex = 8
aeBuildSyntaxUncoercedHex = 9
aeBuildSyntaxNoCloseString = 10
aeBuildSyntaxBadDesc = 11
aeBuildSyntaxBadData = 12
aeBuildSyntaxNoCloseParen = 13
aeBuildSyntaxNoCloseBracket = 14
aeBuildSyntaxNoCloseBrace = 15
aeBuildSyntaxNoKey = 16
aeBuildSyntaxNoColon = 17
aeBuildSyntaxCoercedList = 18
aeBuildSyntaxUncoercedDoubleAt = 19



# AEMach.h

keyReplyPortAttr = 'repp'

typeReplyPortAttr = keyReplyPortAttr



# AEObjects.h

kAEAND = 'AND '
kAEOR = 'OR  '
kAENOT = 'NOT '
kAEFirst = 'firs'
kAELast = 'last'
kAEMiddle = 'midd'
kAEAny = 'any '
kAEAll = 'all '
kAENext = 'next'
kAEPrevious = 'prev'
keyAECompOperator = 'relo'
keyAELogicalTerms = 'term'
keyAELogicalOperator = 'logc'
keyAEObject1 = 'obj1'
keyAEObject2 = 'obj2'
keyAEDesiredClass = 'want'
keyAEContainer = 'from'
keyAEKeyForm = 'form'
keyAEKeyData = 'seld'

keyAERangeStart = 'star'
keyAERangeStop = 'stop'
keyDisposeTokenProc = 'xtok'
keyAECompareProc = 'cmpr'
keyAECountProc = 'cont'
keyAEMarkTokenProc = 'mkid'
keyAEMarkProc = 'mark'
keyAEAdjustMarksProc = 'adjm'
keyAEGetErrDescProc = 'indc'

formAbsolutePosition = 'indx'
formRelativePosition = 'rele'
formTest = 'test'
formRange = 'rang'
formPropertyID = 'prop'
formName = 'name'
formUniqueID = 'ID  '
typeObjectSpecifier = 'obj '
typeObjectBeingExamined = 'exmn'
typeCurrentContainer = 'ccnt'
typeToken = 'toke'
typeRelativeDescriptor = 'rel '
typeAbsoluteOrdinal = 'abso'
typeIndexDescriptor = 'inde'
typeRangeDescriptor = 'rang'
typeLogicalDescriptor = 'logi'
typeCompDescriptor = 'cmpd'
typeOSLTokenList = 'ostl'

kAEIDoMinimum = 0x0000
kAEIDoWhose = 0x0001
kAEIDoMarking = 0x0004
kAEPassSubDescs = 0x0008
kAEResolveNestedLists = 0x0010
kAEHandleSimpleRanges = 0x0020
kAEUseRelativeIterators = 0x0040

typeWhoseDescriptor = 'whos'
formWhose = 'whos'
typeWhoseRange = 'wrng'
keyAEWhoseRangeStart = 'wstr'
keyAEWhoseRangeStop = 'wstp'
keyAEIndex = 'kidx'
keyAETest = 'ktst'



# AEPackObject.h



# AERegistry.h

cAEList = 'list'
cApplication = 'capp'
cArc = 'carc'
cBoolean = 'bool'
cCell = 'ccel'
cChar = 'cha '
cColorTable = 'clrt'
cColumn = 'ccol'
cDocument = 'docu'
cDrawingArea = 'cdrw'
cEnumeration = 'enum'
cFile = 'file'
cFixed = 'fixd'
cFixedPoint = 'fpnt'
cFixedRectangle = 'frct'
cGraphicLine = 'glin'
cGraphicObject = 'cgob'
cGraphicShape = 'cgsh'
cGraphicText = 'cgtx'
cGroupedGraphic = 'cpic'

cInsertionLoc = 'insl'
cInsertionPoint = 'cins'
cIntlText = 'itxt'
cIntlWritingCode = 'intl'
cItem = 'citm'
cLine = 'clin'
cLongDateTime = 'ldt '
cLongFixed = 'lfxd'
cLongFixedPoint = 'lfpt'
cLongFixedRectangle = 'lfrc'
cLongInteger = 'long'
cLongPoint = 'lpnt'
cLongRectangle = 'lrct'
cMachineLoc = 'mLoc'
cMenu = 'cmnu'
cMenuItem = 'cmen'
cObject = 'cobj'
cObjectSpecifier = 'obj '
cOpenableObject = 'coob'
cOval = 'covl'

cParagraph = 'cpar'
cPICT = 'PICT'
cPixel = 'cpxl'
cPixelMap = 'cpix'
cPolygon = 'cpgn'
cProperty = 'prop'
cQDPoint = 'QDpt'
cQDRectangle = 'qdrt'
cRectangle = 'crec'
cRGBColor = 'cRGB'
cRotation = 'trot'
cRoundedRectangle = 'crrc'
cRow = 'crow'
cSelection = 'csel'
cShortInteger = 'shor'
cTable = 'ctbl'
cText = 'ctxt'
cTextFlow = 'cflo'
cTextStyles = 'tsty'
cType = 'type'

cVersion = 'vers'
cWindow = 'cwin'
cWord = 'cwor'
enumArrows = 'arro'
enumJustification = 'just'
enumKeyForm = 'kfrm'
enumPosition = 'posi'
enumProtection = 'prtn'
enumQuality = 'qual'
enumSaveOptions = 'savo'
enumStyle = 'styl'
enumTransferMode = 'tran'
kAEAbout = 'abou'
kAEAfter = 'afte'
kAEAliasSelection = 'sali'
kAEAllCaps = 'alcp'
kAEArrowAtEnd = 'aren'
kAEArrowAtStart = 'arst'
kAEArrowBothEnds = 'arbo'

kAEAsk = 'ask '
kAEBefore = 'befo'
kAEBeginning = 'bgng'
kAEBeginsWith = 'bgwt'
kAEBeginTransaction = 'begi'
kAEBold = 'bold'
kAECaseSensEquals = 'cseq'
kAECentered = 'cent'
kAEChangeView = 'view'
kAEClone = 'clon'
kAEClose = 'clos'
kAECondensed = 'cond'
kAEContains = 'cont'
kAECopy = 'copy'
kAECoreSuite = 'core'
kAECountElements = 'cnte'
kAECreateElement = 'crel'
kAECreatePublisher = 'cpub'
kAECut = 'cut '
kAEDelete = 'delo'

kAEDoObjectsExist = 'doex'
kAEDoScript = 'dosc'
kAEDrag = 'drag'
kAEDuplicateSelection = 'sdup'
kAEEditGraphic = 'edit'
kAEEmptyTrash = 'empt'
kAEEnd = 'end '
kAEEndsWith = 'ends'
kAEEndTransaction = 'endt'
kAEEquals = '=   '
kAEExpanded = 'pexp'
kAEFast = 'fast'
kAEFinderEvents = 'FNDR'
kAEFormulaProtect = 'fpro'
kAEFullyJustified = 'full'
kAEGetClassInfo = 'qobj'
kAEGetData = 'getd'
kAEGetDataSize = 'dsiz'
kAEGetEventInfo = 'gtei'
kAEGetInfoSelection = 'sinf'

kAEGetPrivilegeSelection = 'sprv'
kAEGetSuiteInfo = 'gtsi'
kAEGreaterThan = '>   '
kAEGreaterThanEquals = '>=  '
kAEGrow = 'grow'
kAEHidden = 'hidn'
kAEHiQuality = 'hiqu'
kAEImageGraphic = 'imgr'
kAEIsUniform = 'isun'
kAEItalic = 'ital'
kAELeftJustified = 'left'
kAELessThan = '<   '
kAELessThanEquals = '<=  '
kAELowercase = 'lowc'
kAEMakeObjectsVisible = 'mvis'
kAEMiscStandards = 'misc'
kAEModifiable = 'modf'
kAEMove = 'move'
kAENo = 'no  '
kAENoArrow = 'arno'

kAENonmodifiable = 'nmod'
kAEOpen = 'odoc'
kAEOpenSelection = 'sope'
kAEOutline = 'outl'
kAEPageSetup = 'pgsu'
kAEPaste = 'past'
kAEPlain = 'plan'
kAEPrint = 'pdoc'
kAEPrintSelection = 'spri'
kAEPrintWindow = 'pwin'
kAEPutAwaySelection = 'sput'
kAEQDAddOver = 'addo'
kAEQDAddPin = 'addp'
kAEQDAdMax = 'admx'
kAEQDAdMin = 'admn'
kAEQDBic = 'bic '
kAEQDBlend = 'blnd'
kAEQDCopy = 'cpy '
kAEQDNotBic = 'nbic'
kAEQDNotCopy = 'ncpy'

kAEQDNotOr = 'ntor'
kAEQDNotXor = 'nxor'
kAEQDOr = 'or  '
kAEQDSubOver = 'subo'
kAEQDSubPin = 'subp'
kAEQDSupplementalSuite = 'qdsp'
kAEQDXor = 'xor '
kAEQuickdrawSuite = 'qdrw'
kAEQuitAll = 'quia'
kAERedo = 'redo'
kAERegular = 'regl'
kAEReopenApplication = 'rapp'
kAEReplace = 'rplc'
kAERequiredSuite = 'reqd'
kAERestart = 'rest'
kAERevealSelection = 'srev'
kAERevert = 'rvrt'
kAERightJustified = 'rght'
kAESave = 'save'
kAESelect = 'slct'
kAESetData = 'setd'

kAESetPosition = 'posn'
kAEShadow = 'shad'
kAEShowClipboard = 'shcl'
kAEShutDown = 'shut'
kAESleep = 'slep'
kAESmallCaps = 'smcp'
kAESpecialClassProperties = 'c@#!'
kAEStrikethrough = 'strk'
kAESubscript = 'sbsc'
kAESuperscript = 'spsc'
kAETableSuite = 'tbls'
kAETextSuite = 'TEXT'
kAETransactionTerminated = 'ttrm'
kAEUnderline = 'undl'
kAEUndo = 'undo'
kAEWholeWordEquals = 'wweq'
kAEYes = 'yes '
kAEZoom = 'zoom'

kAELogOut = 'logo'
kAEReallyLogOut = 'rlgo'
kAEShowRestartDialog = 'rrst'
kAEShowShutdownDialog = 'rsdn'

kAEMouseClass = 'mous'
kAEDown = 'down'
kAEUp = 'up  '
kAEMoved = 'move'
kAEStoppedMoving = 'stop'
kAEWindowClass = 'wind'
kAEUpdate = 'updt'
kAEActivate = 'actv'
kAEDeactivate = 'dact'
kAECommandClass = 'cmnd'
kAEKeyClass = 'keyc'
kAERawKey = 'rkey'
kAEVirtualKey = 'keyc'
kAENavigationKey = 'nave'
kAEAutoDown = 'auto'
kAEApplicationClass = 'appl'
kAESuspend = 'susp'
kAEResume = 'rsme'
kAEDiskEvent = 'disk'
kAENullEvent = 'null'
kAEWakeUpEvent = 'wake'
kAEScrapEvent = 'scrp'
kAEHighLevel = 'high'

keyAEAngle = 'kang'
keyAEArcAngle = 'parc'

keyAEBaseAddr = 'badd'
keyAEBestType = 'pbst'
keyAEBgndColor = 'kbcl'
keyAEBgndPattern = 'kbpt'
keyAEBounds = 'pbnd'
keyAECellList = 'kclt'
keyAEClassID = 'clID'
keyAEColor = 'colr'
keyAEColorTable = 'cltb'
keyAECurveHeight = 'kchd'
keyAECurveWidth = 'kcwd'
keyAEDashStyle = 'pdst'
keyAEData = 'data'
keyAEDefaultType = 'deft'
keyAEDefinitionRect = 'pdrt'
keyAEDescType = 'dstp'
keyAEDestination = 'dest'
keyAEDoAntiAlias = 'anta'
keyAEDoDithered = 'gdit'
keyAEDoRotate = 'kdrt'

keyAEDoScale = 'ksca'
keyAEDoTranslate = 'ktra'
keyAEEditionFileLoc = 'eloc'
keyAEElements = 'elms'
keyAEEndPoint = 'pend'
keyAEEventClass = 'evcl'
keyAEEventID = 'evti'
keyAEFile = 'kfil'
keyAEFileType = 'fltp'
keyAEFillColor = 'flcl'
keyAEFillPattern = 'flpt'
keyAEFlipHorizontal = 'kfho'
keyAEFlipVertical = 'kfvt'
keyAEFont = 'font'
keyAEFormula = 'pfor'
keyAEGraphicObjects = 'gobs'
keyAEID = 'ID  '
keyAEImageQuality = 'gqua'
keyAEInsertHere = 'insh'
keyAEKeyForms = 'keyf'

keyAEKeyword = 'kywd'
keyAELevel = 'levl'
keyAELineArrow = 'arro'
keyAEName = 'pnam'
keyAENewElementLoc = 'pnel'
keyAEObject = 'kobj'
keyAEObjectClass = 'kocl'
keyAEOffStyles = 'ofst'
keyAEOnStyles = 'onst'
keyAEParameters = 'prms'
keyAEParamFlags = 'pmfg'
keyAEPenColor = 'ppcl'
keyAEPenPattern = 'pppa'
keyAEPenWidth = 'ppwd'
keyAEPixelDepth = 'pdpt'
keyAEPixMapMinus = 'kpmm'
keyAEPMTable = 'kpmt'
keyAEPointList = 'ptlt'
keyAEPointSize = 'ptsz'
keyAEPosition = 'kpos'

keyAEPropData = 'prdt'
keyAEProperties = 'qpro'
keyAEProperty = 'kprp'
keyAEPropFlags = 'prfg'
keyAEPropID = 'prop'
keyAEProtection = 'ppro'
keyAERenderAs = 'kren'
keyAERequestedType = 'rtyp'
keyAEResult = '----'
keyAEResultInfo = 'rsin'
keyAERotation = 'prot'
keyAERotPoint = 'krtp'
keyAERowList = 'krls'
keyAESaveOptions = 'savo'
keyAEScale = 'pscl'
keyAEScriptTag = 'psct'
keyAESearchText = 'stxt'
keyAEShowWhere = 'show'
keyAEStartAngle = 'pang'
keyAEStartPoint = 'pstp'
keyAEStyles = 'ksty'

keyAESuiteID = 'suit'
keyAEText = 'ktxt'
keyAETextColor = 'ptxc'
keyAETextFont = 'ptxf'
keyAETextPointSize = 'ptps'
keyAETextStyles = 'txst'
keyAETextLineHeight = 'ktlh'
keyAETextLineAscent = 'ktas'
keyAETheText = 'thtx'
keyAETransferMode = 'pptm'
keyAETranslation = 'ptrs'
keyAETryAsStructGraf = 'toog'
keyAEUniformStyles = 'ustl'
keyAEUpdateOn = 'pupd'
keyAEUserTerm = 'utrm'
keyAEWindow = 'wndw'
keyAEWritingCode = 'wrcd'

keyMiscellaneous = 'fmsc'
keySelection = 'fsel'
keyWindow = 'kwnd'
keyWhen = 'when'
keyWhere = 'wher'
keyModifiers = 'mods'
keyKey = 'key '
keyKeyCode = 'code'
keyKeyboard = 'keyb'
keyDriveNumber = 'drv#'
keyErrorCode = 'err#'
keyHighLevelClass = 'hcls'
keyHighLevelID = 'hid '

pArcAngle = 'parc'
pBackgroundColor = 'pbcl'
pBackgroundPattern = 'pbpt'
pBestType = 'pbst'
pBounds = 'pbnd'
pClass = 'pcls'
pClipboard = 'pcli'
pColor = 'colr'
pColorTable = 'cltb'
pContents = 'pcnt'
pCornerCurveHeight = 'pchd'
pCornerCurveWidth = 'pcwd'
pDashStyle = 'pdst'
pDefaultType = 'deft'
pDefinitionRect = 'pdrt'
pEnabled = 'enbl'
pEndPoint = 'pend'
pFillColor = 'flcl'
pFillPattern = 'flpt'
pFont = 'font'

pFormula = 'pfor'
pGraphicObjects = 'gobs'
pHasCloseBox = 'hclb'
pHasTitleBar = 'ptit'
pID = 'ID  '
pIndex = 'pidx'
pInsertionLoc = 'pins'
pIsFloating = 'isfl'
pIsFrontProcess = 'pisf'
pIsModal = 'pmod'
pIsModified = 'imod'
pIsResizable = 'prsz'
pIsStationeryPad = 'pspd'
pIsZoomable = 'iszm'
pIsZoomed = 'pzum'
pItemNumber = 'itmn'
pJustification = 'pjst'
pLineArrow = 'arro'
pMenuID = 'mnid'
pName = 'pnam'

pNewElementLoc = 'pnel'
pPenColor = 'ppcl'
pPenPattern = 'pppa'
pPenWidth = 'ppwd'
pPixelDepth = 'pdpt'
pPointList = 'ptlt'
pPointSize = 'ptsz'
pProtection = 'ppro'
pRotation = 'prot'
pScale = 'pscl'
pScript = 'scpt'
pScriptTag = 'psct'
pSelected = 'selc'
pSelection = 'sele'
pStartAngle = 'pang'
pStartPoint = 'pstp'
pTextColor = 'ptxc'
pTextFont = 'ptxf'
pTextItemDelimiters = 'txdl'
pTextPointSize = 'ptps'

pTextStyles = 'txst'
pTransferMode = 'pptm'
pTranslation = 'ptrs'
pUniformStyles = 'ustl'
pUpdateOn = 'pupd'
pUserSelection = 'pusl'
pVersion = 'vers'
pVisible = 'pvis'

typeAEText = 'tTXT'
typeArc = 'carc'
typeBest = 'best'
typeCell = 'ccel'
typeClassInfo = 'gcli'
typeColorTable = 'clrt'
typeColumn = 'ccol'
typeDashStyle = 'tdas'
typeData = 'tdta'
typeDrawingArea = 'cdrw'
typeElemInfo = 'elin'
typeEnumeration = 'enum'
typeEPS = 'EPS '
typeEventInfo = 'evin'

typeFinderWindow = 'fwin'
typeFixedPoint = 'fpnt'
typeFixedRectangle = 'frct'
typeGraphicLine = 'glin'
typeGraphicText = 'cgtx'
typeGroupedGraphic = 'cpic'
typeInsertionLoc = 'insl'
typeIntlText = 'itxt'
typeIntlWritingCode = 'intl'
typeLongDateTime = 'ldt '
typeCFAbsoluteTime = 'cfat'
typeISO8601DateTime = 'isot'
typeLongFixed = 'lfxd'
typeLongFixedPoint = 'lfpt'
typeLongFixedRectangle = 'lfrc'
typeLongPoint = 'lpnt'
typeLongRectangle = 'lrct'
typeMachineLoc = 'mLoc'
typeOval = 'covl'
typeParamInfo = 'pmin'
typePict = 'PICT'

typePixelMap = 'cpix'
typePixMapMinus = 'tpmm'
typePolygon = 'cpgn'
typePropInfo = 'pinf'
typePtr = 'ptr '
typeQDPoint = 'QDpt'
typeQDRegion = 'Qrgn'
typeRectangle = 'crec'
typeRGB16 = 'tr16'
typeRGB96 = 'tr96'
typeRGBColor = 'cRGB'
typeRotation = 'trot'
typeRoundedRectangle = 'crrc'
typeRow = 'crow'
typeScrapStyles = 'styl'
typeScript = 'scpt'
typeStyledText = 'STXT'
typeSuiteInfo = 'suin'
typeTable = 'ctbl'
typeTextStyles = 'tsty'

typeTIFF = 'TIFF'
typeJPEG = 'JPEG'
typeGIF = 'GIFf'
typeVersion = 'vers'

kAEMenuClass = 'menu'
kAEMenuSelect = 'mhit'
kAEMouseDown = 'mdwn'
kAEMouseDownInBack = 'mdbk'
kAEKeyDown = 'kdwn'
kAEResized = 'rsiz'
kAEPromise = 'prom'

keyMenuID = 'mid '
keyMenuItem = 'mitm'
keyCloseAllWindows = 'caw '
keyOriginalBounds = 'obnd'
keyNewBounds = 'nbnd'
keyLocalWhere = 'lwhr'

typeHIMenu = 'mobj'
typeHIWindow = 'wobj'

kBySmallIcon = 0
kByIconView = 1
kByNameView = 2
kByDateView = 3
kBySizeView = 4
kByKindView = 5
kByCommentView = 6
kByLabelView = 7
kByVersionView = 8

kAEInfo = 11
kAEMain = 0
kAESharing = 13

kAEZoomIn = 7
kAEZoomOut = 8

kTextServiceClass = 'tsvc'
kUpdateActiveInputArea = 'updt'
kShowHideInputWindow = 'shiw'
kPos2Offset = 'p2st'
kOffset2Pos = 'st2p'
kUnicodeNotFromInputMethod = 'unim'
kGetSelectedText = 'gtxt'
keyAETSMDocumentRefcon = 'refc'
keyAEServerInstance = 'srvi'
keyAETheData = 'kdat'
keyAEFixLength = 'fixl'
keyAEUpdateRange = 'udng'
keyAECurrentPoint = 'cpos'
keyAEBufferSize = 'buff'
keyAEMoveView = 'mvvw'
keyAENextBody = 'nxbd'
keyAETSMScriptTag = 'sclg'
keyAETSMTextFont = 'ktxf'
keyAETSMTextFMFont = 'ktxm'
keyAETSMTextPointSize = 'ktps'
keyAETSMEventRecord = 'tevt'
keyAETSMEventRef = 'tevr'
keyAETextServiceEncoding = 'tsen'
keyAETextServiceMacEncoding = 'tmen'
keyAETSMGlyphInfoArray = 'tgia'
typeTextRange = 'txrn'
typeComponentInstance = 'cmpi'
typeOffsetArray = 'ofay'
typeTextRangeArray = 'tray'
typeLowLevelEventRecord = 'evtr'
typeGlyphInfoArray = 'glia'
typeEventRef = 'evrf'
typeText = typeChar

kTSMOutsideOfBody = 1
kTSMInsideOfBody = 2
kTSMInsideOfActiveInputArea = 3

kNextBody = 1
kPreviousBody = 2

kTSMHiliteCaretPosition = 1
kTSMHiliteRawText = 2
kTSMHiliteSelectedRawText = 3
kTSMHiliteConvertedText = 4
kTSMHiliteSelectedConvertedText = 5
kTSMHiliteBlockFillText = 6
kTSMHiliteOutlineText = 7
kTSMHiliteSelectedText = 8
kTSMHiliteNoHilite = 9

kCaretPosition = kTSMHiliteCaretPosition
kRawText = kTSMHiliteRawText
kSelectedRawText = kTSMHiliteSelectedRawText
kConvertedText = kTSMHiliteConvertedText
kSelectedConvertedText = kTSMHiliteSelectedConvertedText
kBlockFillText = kTSMHiliteBlockFillText
kOutlineText = kTSMHiliteOutlineText
kSelectedText = kTSMHiliteSelectedText

keyAEHiliteRange = 'hrng'
keyAEPinRange = 'pnrg'
keyAEClauseOffsets = 'clau'
keyAEOffset = 'ofst'
keyAEPoint = 'gpos'
keyAELeftSide = 'klef'
keyAERegionClass = 'rgnc'
keyAEDragging = 'bool'

keyAELeadingEdge = keyAELeftSide

typeMeters = 'metr'
typeInches = 'inch'
typeFeet = 'feet'
typeYards = 'yard'
typeMiles = 'mile'
typeKilometers = 'kmtr'
typeCentimeters = 'cmtr'
typeSquareMeters = 'sqrm'
typeSquareFeet = 'sqft'
typeSquareYards = 'sqyd'
typeSquareMiles = 'sqmi'
typeSquareKilometers = 'sqkm'
typeLiters = 'litr'
typeQuarts = 'qrts'
typeGallons = 'galn'
typeCubicMeters = 'cmet'
typeCubicFeet = 'cfet'
typeCubicInches = 'cuin'
typeCubicCentimeter = 'ccmt'
typeCubicYards = 'cyrd'
typeKilograms = 'kgrm'
typeGrams = 'gram'
typeOunces = 'ozs '
typePounds = 'lbs '
typeDegreesC = 'degc'
typeDegreesF = 'degf'
typeDegreesK = 'degk'

kFAServerApp = 'ssrv'
kDoFolderActionEvent = 'fola'
kFolderActionCode = 'actn'
kFolderOpenedEvent = 'fopn'
kFolderClosedEvent = 'fclo'
kFolderWindowMovedEvent = 'fsiz'
kFolderItemsAddedEvent = 'fget'
kFolderItemsRemovedEvent = 'flos'
kItemList = 'flst'
kNewSizeParameter = 'fnsz'
kFASuiteCode = 'faco'
kFAAttachCommand = 'atfa'
kFARemoveCommand = 'rmfa'
kFAEditCommand = 'edfa'
kFAFileParam = 'faal'
kFAIndexParam = 'indx'

kAEInternetSuite = 'gurl'
kAEISWebStarSuite = 'WWW\xBD'

kAEISGetURL = 'gurl'
KAEISHandleCGI = 'sdoc'

cURL = 'url '
cInternetAddress = 'IPAD'
cHTML = 'html'
cFTPItem = 'ftp '

kAEISHTTPSearchArgs = 'kfor'
kAEISPostArgs = 'post'
kAEISMethod = 'meth'
kAEISClientAddress = 'addr'
kAEISUserName = 'user'
kAEISPassword = 'pass'
kAEISFromUser = 'frmu'
kAEISServerName = 'svnm'
kAEISServerPort = 'svpt'
kAEISScriptName = 'scnm'
kAEISContentType = 'ctyp'
kAEISReferrer = 'refr'
kAEISUserAgent = 'Agnt'
kAEISAction = 'Kact'
kAEISActionPath = 'Kapt'
kAEISClientIP = 'Kcip'
kAEISFullRequest = 'Kfrq'

pScheme = 'pusc'
pHost = 'HOST'
pPath = 'FTPc'
pUserName = 'RAun'
pUserPassword = 'RApw'
pDNSForm = 'pDNS'
pURL = 'pURL'
pTextEncoding = 'ptxe'
pFTPKind = 'kind'

eScheme = 'esch'
eurlHTTP = 'http'
eurlHTTPS = 'htps'
eurlFTP = 'ftp '
eurlMail = 'mail'
eurlFile = 'file'
eurlGopher = 'gphr'
eurlTelnet = 'tlnt'
eurlNews = 'news'
eurlSNews = 'snws'
eurlNNTP = 'nntp'
eurlMessage = 'mess'
eurlMailbox = 'mbox'
eurlMulti = 'mult'
eurlLaunch = 'laun'
eurlAFP = 'afp '
eurlAT = 'at  '
eurlEPPC = 'eppc'
eurlRTSP = 'rtsp'
eurlIMAP = 'imap'
eurlNFS = 'unfs'
eurlPOP = 'upop'
eurlLDAP = 'uldp'
eurlUnknown = 'url?'

kConnSuite = 'macc'
cDevSpec = 'cdev'
cAddressSpec = 'cadr'
cADBAddress = 'cadb'
cAppleTalkAddress = 'cat '
cBusAddress = 'cbus'
cEthernetAddress = 'cen '
cFireWireAddress = 'cfw '
cIPAddress = 'cip '
cLocalTalkAddress = 'clt '
cSCSIAddress = 'cscs'
cTokenRingAddress = 'ctok'
cUSBAddress = 'cusb'
pDeviceType = 'pdvt'
pDeviceAddress = 'pdva'
pConduit = 'pcon'
pProtocol = 'pprt'
pATMachine = 'patm'
pATZone = 'patz'
pATType = 'patt'
pDottedDecimal = 'pipd'
pDNS = 'pdns'
pPort = 'ppor'
pNetwork = 'pnet'
pNode = 'pnod'
pSocket = 'psoc'
pSCSIBus = 'pscb'
pSCSILUN = 'pslu'
eDeviceType = 'edvt'
eAddressSpec = 'eads'
eConduit = 'econ'
eProtocol = 'epro'
eADB = 'eadb'
eAnalogAudio = 'epau'
eAppleTalk = 'epat'
eAudioLineIn = 'ecai'
eAudioLineOut = 'ecal'
eAudioOut = 'ecao'
eBus = 'ebus'
eCDROM = 'ecd '
eCommSlot = 'eccm'
eDigitalAudio = 'epda'
eDisplay = 'edds'
eDVD = 'edvd'
eEthernet = 'ecen'
eFireWire = 'ecfw'
eFloppy = 'efd '
eHD = 'ehd '
eInfrared = 'ecir'
eIP = 'epip'
eIrDA = 'epir'
eIRTalk = 'epit'
eKeyboard = 'ekbd'
eLCD = 'edlc'
eLocalTalk = 'eclt'
eMacIP = 'epmi'
eMacVideo = 'epmv'
eMicrophone = 'ecmi'
eModemPort = 'ecmp'
eModemPrinterPort = 'empp'
eModem = 'edmm'
eMonitorOut = 'ecmn'
eMouse = 'emou'
eNuBusCard = 'ednb'
eNuBus = 'enub'
ePCcard = 'ecpc'
ePCIbus = 'ecpi'
ePCIcard = 'edpi'
ePDSslot = 'ecpd'
ePDScard = 'epds'
ePointingDevice = 'edpd'
ePostScript = 'epps'
ePPP = 'eppp'
ePrinterPort = 'ecpp'
ePrinter = 'edpr'
eSvideo = 'epsv'
eSCSI = 'ecsc'
eSerial = 'epsr'
eSpeakers = 'edsp'
eStorageDevice = 'edst'
eSVGA = 'epsg'
eTokenRing = 'etok'
eTrackball = 'etrk'
eTrackpad = 'edtp'
eUSB = 'ecus'
eVideoIn = 'ecvi'
eVideoMonitor = 'edvm'
eVideoOut = 'ecvo'

cKeystroke = 'kprs'
pKeystrokeKey = 'kMsg'
pModifiers = 'kMod'
pKeyKind = 'kknd'
eModifiers = 'eMds'
eOptionDown = 'Kopt'
eCommandDown = 'Kcmd'
eControlDown = 'Kctl'
eShiftDown = 'Ksft'
eCapsLockDown = 'Kclk'
eKeyKind = 'ekst'
eEscapeKey = 'ks5\x00'
eDeleteKey = 'ks3\x00'
eTabKey = 'ks0\x00'
eReturnKey = 'ks\x24\x00'
eClearKey = 'ksG\x00'
eEnterKey = 'ksL\x00'
eUpArrowKey = 'ks\x7E\x00'
eDownArrowKey = 'ks\x7D\x00'
eLeftArrowKey = 'ks\x7B\x00'
eRightArrowKey = 'ks\x7C\x00'
eHelpKey = 'ksr\x00'
eHomeKey = 'kss\x00'
ePageUpKey = 'kst\x00'
ePageDownKey = 'ksy\x00'
eForwardDelKey = 'ksu\x00'
eEndKey = 'ksw\x00'
eF1Key = 'ksz\x00'
eF2Key = 'ksx\x00'
eF3Key = 'ksc\x00'
eF4Key = 'ksv\x00'
eF5Key = 'ks\x60\x00'
eF6Key = 'ksa\x00'
eF7Key = 'ksb\x00'
eF8Key = 'ksd\x00'
eF9Key = 'kse\x00'
eF10Key = 'ksm\x00'
eF11Key = 'ksg\x00'
eF12Key = 'kso\x00'
eF13Key = 'ksi\x00'
eF14Key = 'ksk\x00'
eF15Key = 'ksq\x00'

keyAELaunchedAsLogInItem = 'lgit'
keyAELaunchedAsServiceItem = 'svit'



# AEUserTermTypes.h

kAEUserTerminology = 'aeut'
kAETerminologyExtension = 'aete'
kAEScriptingSizeResource = 'scsz'
kAEOSAXSizeResource = 'osiz'

kAEUTHasReturningParam = 31
kAEUTOptional = 15
kAEUTlistOfItems = 14
kAEUTEnumerated = 13
kAEUTReadWrite = 12
kAEUTChangesState = 12
kAEUTTightBindingFunction = 12
kAEUTEnumsAreTypes = 11
kAEUTEnumListIsExclusive = 10
kAEUTReplyIsReference = 9
kAEUTDirectParamIsReference = 9
kAEUTParamIsReference = 9
kAEUTPropertyIsReference = 9
kAEUTNotDirectParamIsTarget = 8
kAEUTParamIsTarget = 8
kAEUTApostrophe = 3
kAEUTFeminine = 2
kAEUTMasculine = 1
kAEUTPlural = 0

kLaunchToGetTerminology = (1 << 15)
kDontFindAppBySignature = (1 << 14)
kAlwaysSendSubject = (1 << 13)

kReadExtensionTermsMask = (1 << 15)

kOSIZDontOpenResourceFile = 15
kOSIZdontAcceptRemoteEvents = 14
kOSIZOpenWithReadPermission = 13
kOSIZCodeInSharedLibraries = 11



# AppleEvents.h

keyDirectObject = '----'
keyErrorNumber = 'errn'
keyErrorString = 'errs'
keyProcessSerialNumber = 'psn '
keyPreDispatch = 'phac'
keySelectProc = 'selh'
keyAERecorderCount = 'recr'
keyAEVersion = 'vers'

kCoreEventClass = 'aevt'

kAEOpenApplication = 'oapp'
kAEOpenDocuments = 'odoc'
kAEPrintDocuments = 'pdoc'
kAEOpenContents = 'ocon'
kAEQuitApplication = 'quit'
kAEAnswer = 'ansr'
kAEApplicationDied = 'obit'
kAEShowPreferences = 'pref'

kAEStartRecording = 'reca'
kAEStopRecording = 'recc'
kAENotifyStartRecording = 'rec1'
kAENotifyStopRecording = 'rec0'
kAENotifyRecording = 'recr'

kAEUnknownSource = 0
kAEDirectCall = 1
kAESameProcess = 2
kAELocalProcess = 3
kAERemoteProcess = 4



# AEInteraction.h

kAEInteractWithSelf = 0
kAEInteractWithLocal = 1
kAEInteractWithAll = 2

kAEDoNotIgnoreHandler = 0x00000000
kAEIgnoreAppPhacHandler = 0x00000001
kAEIgnoreAppEventHandler = 0x00000002
kAEIgnoreSysPhacHandler = 0x00000004
kAEIgnoreSysEventHandler = 0x00000008
kAEIngoreBuiltInEventHandler = 0x00000010
kAEDontDisposeOnResume = 0x80000000

kAENoDispatch = 0
kAEUseStandardDispatch = 0xFFFFFFFF



# AppleScript.h

typeAppleScript = 'ascr'
kAppleScriptSubtype = typeAppleScript
typeASStorage = typeAppleScript

kASSelectInit = 0x1001
kASSelectSetSourceStyles = 0x1002
kASSelectGetSourceStyles = 0x1003
kASSelectGetSourceStyleNames = 0x1004
kASSelectCopySourceAttributes = 0x1005
kASSelectSetSourceAttributes = 0x1006

kASHasOpenHandler = 'hsod'

kASDefaultMinStackSize = 4
kASDefaultPreferredStackSize = 16
kASDefaultMaxStackSize = 16
kASDefaultMinHeapSize = 4
kASDefaultPreferredHeapSize = 16
kASDefaultMaxHeapSize = 32L

kASSourceStyleUncompiledText = 0
kASSourceStyleNormalText = 1
kASSourceStyleLanguageKeyword = 2
kASSourceStyleApplicationKeyword = 3
kASSourceStyleComment = 4
kASSourceStyleLiteral = 5
kASSourceStyleUserSymbol = 6
kASSourceStyleObjectSpecifier = 7
kASNumberOfSourceStyles = 8



# ASDebugging.h

kOSAModeDontDefine = 0x0001

kASSelectSetPropertyObsolete = 0x1101
kASSelectGetPropertyObsolete = 0x1102
kASSelectSetHandlerObsolete = 0x1103
kASSelectGetHandlerObsolete = 0x1104
kASSelectGetAppTerminologyObsolete = 0x1105
kASSelectSetProperty = 0x1106
kASSelectGetProperty = 0x1107
kASSelectSetHandler = 0x1108
kASSelectGetHandler = 0x1109
kASSelectGetAppTerminology = 0x110A
kASSelectGetSysTerminology = 0x110B
kASSelectGetPropertyNames = 0x110C
kASSelectGetHandlerNames = 0x110D



# ASRegistry.h

keyAETarget = 'targ'
keySubjectAttr = 'subj'
keyASReturning = 'Krtn'
kASAppleScriptSuite = 'ascr'
kASScriptEditorSuite = 'ToyS'
kASTypeNamesSuite = 'tpnm'
typeAETE = 'aete'
typeAEUT = 'aeut'
kGetAETE = 'gdte'
kGetAEUT = 'gdut'
kUpdateAEUT = 'udut'
kUpdateAETE = 'udte'
kCleanUpAEUT = 'cdut'
kASComment = 'cmnt'
kASLaunchEvent = 'noop'
keyScszResource = 'scsz'
typeScszResource = 'scsz'
kASSubroutineEvent = 'psbr'
keyASSubroutineName = 'snam'
kASPrepositionalSubroutine = 'psbr'
keyASPositionalArgs = 'parg'

keyAppHandledCoercion = 'idas'

kASStartLogEvent = 'log1'
kASStopLogEvent = 'log0'
kASCommentEvent = 'cmnt'

kASAdd = '+   '
kASSubtract = '-   '
kASMultiply = '*   '
kASDivide = '/   '
kASQuotient = 'div '
kASRemainder = 'mod '
kASPower = '^   '
kASEqual = kAEEquals
kASNotEqual = 0xAD202020
kASGreaterThan = kAEGreaterThan
kASGreaterThanOrEqual = kAEGreaterThanEquals
kASLessThan = kAELessThan
kASLessThanOrEqual = kAELessThanEquals
kASComesBefore = 'cbfr'
kASComesAfter = 'cafr'
kASConcatenate = 'ccat'
kASStartsWith = kAEBeginsWith
kASEndsWith = kAEEndsWith
kASContains = kAEContains

kASAnd = kAEAND
kASOr = kAEOR
kASNot = kAENOT
kASNegate = 'neg '
keyASArg = 'arg '

kASErrorEventCode = 'err '
kOSAErrorArgs = 'erra'
keyAEErrorObject = 'erob'
pLength = 'leng'
pReverse = 'rvse'
pRest = 'rest'
pInherits = 'c@#^'
pProperties = 'pALL'
keyASUserRecordFields = 'usrf'
typeUserRecordFields = typeAEList

keyASPrepositionAt = 'at  '
keyASPrepositionIn = 'in  '
keyASPrepositionFrom = 'from'
keyASPrepositionFor = 'for '
keyASPrepositionTo = 'to  '
keyASPrepositionThru = 'thru'
keyASPrepositionThrough = 'thgh'
keyASPrepositionBy = 'by  '
keyASPrepositionOn = 'on  '
keyASPrepositionInto = 'into'
keyASPrepositionOnto = 'onto'
keyASPrepositionBetween = 'btwn'
keyASPrepositionAgainst = 'agst'
keyASPrepositionOutOf = 'outo'
keyASPrepositionInsteadOf = 'isto'
keyASPrepositionAsideFrom = 'asdf'
keyASPrepositionAround = 'arnd'
keyASPrepositionBeside = 'bsid'
keyASPrepositionBeneath = 'bnth'
keyASPrepositionUnder = 'undr'

keyASPrepositionOver = 'over'
keyASPrepositionAbove = 'abve'
keyASPrepositionBelow = 'belw'
keyASPrepositionApartFrom = 'aprt'
keyASPrepositionGiven = 'givn'
keyASPrepositionWith = 'with'
keyASPrepositionWithout = 'wout'
keyASPrepositionAbout = 'abou'
keyASPrepositionSince = 'snce'
keyASPrepositionUntil = 'till'

kDialectBundleResType = 'Dbdl'
cConstant = typeEnumerated
cClassIdentifier = pClass
cObjectBeingExamined = typeObjectBeingExamined
cList = typeAEList
cSmallReal = typeIEEE32BitFloatingPoint
cReal = typeIEEE64BitFloatingPoint
cRecord = typeAERecord
cReference = cObjectSpecifier
cUndefined = 'undf'
cMissingValue = 'msng'
cSymbol = 'symb'
cLinkedList = 'llst'
cVector = 'vect'
cEventIdentifier = 'evnt'
cKeyIdentifier = 'kyid'
cUserIdentifier = 'uid '
cPreposition = 'prep'
cKeyForm = enumKeyForm
cScript = 'scpt'
cHandler = 'hand'
cProcedure = 'proc'

cHandleBreakpoint = 'brak'

cClosure = 'clsr'
cRawData = 'rdat'
cStringClass = typeChar
cNumber = 'nmbr'
cListElement = 'celm'
cListOrRecord = 'lr  '
cListOrString = 'ls  '
cListRecordOrString = 'lrs '
cNumberOrString = 'ns  '
cNumberOrDateTime = 'nd  '
cNumberDateTimeOrString = 'nds '
cAliasOrString = 'sf  '
cSeconds = 'scnd'
typeSound = 'snd '
enumBooleanValues = 'boov'
kAETrue = typeTrue
kAEFalse = typeFalse
enumMiscValues = 'misc'
kASCurrentApplication = 'cura'
formUserPropertyID = 'usrp'

cString = cStringClass

pASIt = 'it  '
pASMe = 'me  '
pASResult = 'rslt'
pASSpace = 'spac'
pASReturn = 'ret '
pASTab = 'tab '
pASPi = 'pi  '
pASParent = 'pare'
kASInitializeEventCode = 'init'
pASPrintLength = 'prln'
pASPrintDepth = 'prdp'
pASTopLevelScript = 'ascr'

kAECase = 'case'
kAEDiacritic = 'diac'
kAEWhiteSpace = 'whit'
kAEHyphens = 'hyph'
kAEExpansion = 'expa'
kAEPunctuation = 'punc'
kAEZenkakuHankaku = 'zkhk'
kAESmallKana = 'skna'
kAEKataHiragana = 'hika'
kASConsiderReplies = 'rmte'
kASNumericStrings = 'nume'
enumConsiderations = 'cons'

kAECaseConsiderMask = 0x00000001
kAEDiacriticConsiderMask = 0x00000002
kAEWhiteSpaceConsiderMask = 0x00000004
kAEHyphensConsiderMask = 0x00000008
kAEExpansionConsiderMask = 0x00000010
kAEPunctuationConsiderMask = 0x00000020
kASConsiderRepliesConsiderMask = 0x00000040
kASNumericStringsConsiderMask = 0x00000080
kAECaseIgnoreMask = 0x00010000
kAEDiacriticIgnoreMask = 0x00020000
kAEWhiteSpaceIgnoreMask = 0x00040000
kAEHyphensIgnoreMask = 0x00080000
kAEExpansionIgnoreMask = 0x00100000
kAEPunctuationIgnoreMask = 0x00200000
kASConsiderRepliesIgnoreMask = 0x00400000
kASNumericStringsIgnoreMask = 0x00800000
enumConsidsAndIgnores = 'csig'

cCoercion = 'coec'
cCoerceUpperCase = 'txup'
cCoerceLowerCase = 'txlo'
cCoerceRemoveDiacriticals = 'txdc'
cCoerceRemovePunctuation = 'txpc'
cCoerceRemoveHyphens = 'txhy'
cCoerceOneByteToTwoByte = 'txex'
cCoerceRemoveWhiteSpace = 'txws'
cCoerceSmallKana = 'txsk'
cCoerceZenkakuhankaku = 'txze'
cCoerceKataHiragana = 'txkh'
cZone = 'zone'
cMachine = 'mach'
cAddress = 'addr'
cRunningAddress = 'radd'
cStorage = 'stor'

pASWeekday = 'wkdy'
pASMonth = 'mnth'
pASDay = 'day '
pASYear = 'year'
pASTime = 'time'
pASDateString = 'dstr'
pASTimeString = 'tstr'
cMonth = pASMonth
cJanuary = 'jan '
cFebruary = 'feb '
cMarch = 'mar '
cApril = 'apr '
cMay = 'may '
cJune = 'jun '
cJuly = 'jul '
cAugust = 'aug '
cSeptember = 'sep '
cOctober = 'oct '
cNovember = 'nov '
cDecember = 'dec '

cWeekday = pASWeekday
cSunday = 'sun '
cMonday = 'mon '
cTuesday = 'tue '
cWednesday = 'wed '
cThursday = 'thu '
cFriday = 'fri '
cSaturday = 'sat '
pASQuote = 'quot'
pASSeconds = 'secs'
pASMinutes = 'min '
pASHours = 'hour'
pASDays = 'days'
pASWeeks = 'week'
cWritingCodeInfo = 'citl'
pScriptCode = 'pscd'
pLangCode = 'plcd'
kASMagicTellEvent = 'tell'
kASMagicEndTellEvent = 'tend'



# DigitalHubRegistry.h

kDigiHubEventClass = 'dhub'

kDigiHubMusicCD = 'aucd'
kDigiHubPictureCD = 'picd'
kDigiHubVideoDVD = 'vdvd'
kDigiHubBlankCD = 'bcd '
kDigiHubBlankDVD = 'bdvd'



# OSA.h

kOSAComponentType = 'osa '

kOSAGenericScriptingComponentSubtype = 'scpt'

kOSAFileType = 'osas'

kOSASuite = 'ascr'

kOSARecordedText = 'recd'

kOSAScriptIsModified = 'modi'

kOSAScriptIsTypeCompiledScript = 'cscr'

kOSAScriptIsTypeScriptValue = 'valu'

kOSAScriptIsTypeScriptContext = 'cntx'

kOSAScriptBestType = 'best'

kOSACanGetSource = 'gsrc'

typeOSADialectInfo = 'difo'
keyOSADialectName = 'dnam'
keyOSADialectCode = 'dcod'
keyOSADialectLangCode = 'dlcd'
keyOSADialectScriptCode = 'dscd'

kOSANullScript = 0

kOSANullMode = 0
kOSAModeNull = 0

kOSASupportsCompiling = 0x0002
kOSASupportsGetSource = 0x0004
kOSASupportsAECoercion = 0x0008
kOSASupportsAESending = 0x0010
kOSASupportsRecording = 0x0020
kOSASupportsConvenience = 0x0040
kOSASupportsDialects = 0x0080
kOSASupportsEventHandling = 0x0100

kOSASelectLoad = 0x0001
kOSASelectStore = 0x0002
kOSASelectExecute = 0x0003
kOSASelectDisplay = 0x0004
kOSASelectScriptError = 0x0005
kOSASelectDispose = 0x0006
kOSASelectSetScriptInfo = 0x0007
kOSASelectGetScriptInfo = 0x0008
kOSASelectSetActiveProc = 0x0009
kOSASelectGetActiveProc = 0x000A
kOSASelectCopyDisplayString = 0x000B

kOSASelectScriptingComponentName = 0x0102
kOSASelectCompile = 0x0103
kOSASelectCopyID = 0x0104

kOSASelectCopyScript = 0x0105

kOSASelectGetSource = 0x0201
kOSASelectCopySourceString = 0x0202

kOSASelectCoerceFromDesc = 0x0301
kOSASelectCoerceToDesc = 0x0302

kOSASelectSetSendProc = 0x0401
kOSASelectGetSendProc = 0x0402
kOSASelectSetCreateProc = 0x0403
kOSASelectGetCreateProc = 0x0404
kOSASelectSetDefaultTarget = 0x0405

kOSASelectStartRecording = 0x0501
kOSASelectStopRecording = 0x0502

kOSASelectLoadExecute = 0x0601
kOSASelectCompileExecute = 0x0602
kOSASelectDoScript = 0x0603

kOSASelectSetCurrentDialect = 0x0701
kOSASelectGetCurrentDialect = 0x0702
kOSASelectAvailableDialects = 0x0703
kOSASelectGetDialectInfo = 0x0704
kOSASelectAvailableDialectCodeList = 0x0705

kOSASelectSetResumeDispatchProc = 0x0801
kOSASelectGetResumeDispatchProc = 0x0802
kOSASelectExecuteEvent = 0x0803
kOSASelectDoEvent = 0x0804
kOSASelectMakeContext = 0x0805

kOSASelectComponentSpecificStart = 0x1001

kOSAModePreventGetSource = 0x00000001

kOSAModeNeverInteract = kAENeverInteract
kOSAModeCanInteract = kAECanInteract
kOSAModeAlwaysInteract = kAEAlwaysInteract
kOSAModeDontReconnect = kAEDontReconnect

kOSAModeCantSwitchLayer = 0x00000040

kOSAModeDoRecord = 0x00001000

kOSAModeCompileIntoContext = 0x00000002

kOSAModeAugmentContext = 0x00000004

kOSAModeDisplayForHumans = 0x00000008

kOSAModeDontStoreParent = 0x00010000

kOSAModeDispatchToDirectObject = 0x00020000

kOSAModeDontGetDataForArguments = 0x00040000

kOSAModeFullyQualifyDescriptors = 0x00080000

kOSAScriptResourceType = kOSAGenericScriptingComponentSubtype

typeOSAGenericStorage = kOSAScriptResourceType

kOSAErrorNumber = keyErrorNumber

kOSAErrorMessage = keyErrorString

kOSAErrorBriefMessage = 'errb'

kOSAErrorApp = 'erap'

kOSAErrorPartialResult = 'ptlr'

kOSAErrorOffendingObject = 'erob'

kOSAErrorExpectedType = 'errt'

kOSAErrorRange = 'erng'

typeOSAErrorRange = 'erng'

keyOSASourceStart = 'srcs'

keyOSASourceEnd = 'srce'

kOSAUseStandardDispatch = kAEUseStandardDispatch

kOSANoDispatch = kAENoDispatch

kOSADontUsePhac = 0x0001



# OSAComp.h



# OSAGeneric.h

kGenericComponentVersion = 0x0100

kGSSSelectGetDefaultScriptingComponent = 0x1001
kGSSSelectSetDefaultScriptingComponent = 0x1002
kGSSSelectGetScriptingComponent = 0x1003
kGSSSelectGetScriptingComponentFromStored = 0x1004
kGSSSelectGenericToRealID = 0x1005
kGSSSelectRealToGenericID = 0x1006
kGSSSelectOutOfRange = 0x1007



# Miscellaneous




"""findapp -- Support module for obtaining the full path to a local application given its name, bundle id or creator type. If application isn't found, an ApplicationNotFoundError exception is raised. """

from os.path import exists

from ae import findapplicationforinfo, MacOSError

__all__ = ['byname', 'byid', 'bycreator']

######################################################################
# PRIVATE
######################################################################

def _findapp(name=None, id=None, creator='????'):
	try:
		return findapplicationforinfo(creator, id, name)
	except MacOSError, err:
		if err.args[0] == -10814:
			raise ApplicationNotFoundError(name or id or creator)
		else:
			raise


######################################################################
# PUBLIC
######################################################################

class ApplicationNotFoundError(Exception):
	def __init__(self, name):
		self.name = name
		Exception.__init__(self, name)
	
	def __str__(self):
		return 'Local application %r not found.' % self.name


def byname(name):
	"""Find the application with the given name and return its full path. 
	
	Absolute paths are also accepted. An '.app' suffix is optional.
	
	Examples: 
		byname('TextEdit')
		byname('Finder.app')
	"""
	if not name.startswith('/'): # application name only, not its full path
		try:
			name = _findapp(name)
		except ApplicationNotFoundError:
			if name.lower().endswith('.app'):
				raise
			name = _findapp(name + '.app')
	if not exists(name) and not name.lower().endswith('.app') and exists(name + '.app'):
		name += '.app'
	if not exists(name):
		raise ApplicationNotFoundError(name)
	return name

		
def byid(id):
	"""Find the application with the given bundle id and return its full path.
	
	Examples:
		byid('com.apple.textedit')
	"""
	return _findapp(id=id)


def bycreator(creator):
	"""Find the application with the given creator type and return its full path.
	
	Examples:
		bycreator('ttxt')
	"""
	if len(creator) != 4 or creator == '????':
		raise ApplicationNotFoundError(creator)
	return _findapp(creator=creator)






"""defaultterminology -- translation tables between appscript-style typenames and corresponding AE codes """

types = [
		# Human-readable names for commonly used AE types.
		# Most of these names are equivalent to AS names, though
		# a few are adjusted to be more 'programmer friendly', 
		# e.g. 'float' instead of 'real', and a few have no AS equivalent,
		# e.g. 'utf8_text'
		('anything', '****'),
		
		('boolean', 'bool'),
		
		('short_integer', 'shor'),
		('integer', 'long'),
		('unsigned_integer', 'magn'),
		('double_integer', 'comp'),
		
		('fixed', 'fixd'),
		('long_fixed', 'lfxd'),
		('decimal_struct', 'decm'),
		
		('short_float', 'sing'),
		('float', 'doub'),
		('extended_float', 'exte'),
		('float_128bit', 'ldbl'),
		
		('string', 'TEXT'),
		('styled_text', 'STXT'),
		('text_style_info', 'tsty'),
		('styled_clipboard_text', 'styl'),
		('encoded_string', 'encs'),
		('writing_code', 'psct'),
		('international_writing_code', 'intl'),
		('international_text', 'itxt'),
		('unicode_text', 'utxt'),
  		('utf8_text', 'utf8'), # typeUTF8Text
		('utf16_text', 'ut16'), # typeUTF16ExternalRepresentation
		
		('version', 'vers'),
		('date', 'ldt '),
		('list', 'list'),
		('record', 'reco'),
		('data', 'tdta'),
		('script', 'scpt'),
		
		('location_reference', 'insl'),
		('reference', 'obj '),
		
		('alias', 'alis'),
		('file_ref', 'fsrf'),
		('file_specification', 'fss '),
		('file_url', 'furl'),
		
		('point', 'QDpt'),
		('bounding_rectangle', 'qdrt'),
		('fixed_point', 'fpnt'),
		('fixed_rectangle', 'frct'),
		('long_point', 'lpnt'),
		('long_rectangle', 'lrct'),
		('long_fixed_point', 'lfpt'),
		('long_fixed_rectangle', 'lfrc'),
		
		('EPS_picture', 'EPS '),
		('GIF_picture', 'GIFf'),
		('JPEG_picture', 'JPEG'),
		('PICT_picture', 'PICT'),
		('TIFF_picture', 'TIFF'),
		('RGB_color', 'cRGB'),
		('RGB16_color', 'tr16'),
		('RGB96_color', 'tr96'),
		('graphic_text', 'cgtx'),
		('color_table', 'clrt'),
		('pixel_map_record', 'tpmm'),
		
		('best', 'best'),
		('type_class', 'type'),
		('enumerator', 'enum'),
		('property', 'prop'),
		
		# AEAddressDesc types
		
		('mach_port', 'port'),
		('kernel_process_id', 'kpid'),
		('application_bundle_id', 'bund'),
		('process_serial_number', 'psn '),
		('application_signature', 'sign'),
		('application_url', 'aprl'),
		
		# misc.
		
		('missing_value', 'msng'),
		
		('null', 'null'),
		
		('machine_location', 'mLoc'),
		('machine', 'mach'),
		
		('dash_style', 'tdas'),
		('rotation', 'trot'),
		
		('suite_info', 'suin'),
		('class_info', 'gcli'),
		('property_info', 'pinf'),
		('element_info', 'elin'),
		('event_info', 'evin'),
		('parameter_info', 'pmin'),
		
		('item', 'cobj'), # Apple have removed the 'item' class definition from skeleton.sdef
		
		# unit types
		
		('centimeters', 'cmtr'),
		('meters', 'metr'),
		('kilometers', 'kmtr'),
		('inches', 'inch'),
		('feet', 'feet'),
		('yards', 'yard'),
		('miles', 'mile'),
		
		('square_meters', 'sqrm'),
		('square_kilometers', 'sqkm'),
		('square_feet', 'sqft'),
		('square_yards', 'sqyd'),
		('square_miles', 'sqmi'),
		
		('cubic_centimeters', 'ccmt'),
		('cubic_meters', 'cmet'),
		('cubic_inches', 'cuin'),
		('cubic_feet', 'cfet'),
		('cubic_yards', 'cyrd'),
		
		('liters', 'litr'),
		('quarts', 'qrts'),
		('gallons', 'galn'),
		
		('grams', 'gram'),
		('kilograms', 'kgrm'),
		('ounces', 'ozs '),
		('pounds', 'lbs '),
		
		('degrees_Celsius', 'degc'),
		('degrees_Fahrenheit', 'degf'),
		('degrees_Kelvin', 'degk'),
		
		# month and weekday
		
		('January', 'jan '),
		('February', 'feb '),
		('March', 'mar '),
		('April', 'apr '),
		('May', 'may '),
		('June', 'jun '),
		('July', 'jul '),
		('August', 'aug '),
		('September', 'sep '),
		('October', 'oct '),
		('November', 'nov '),
		('December', 'dec '),
		
		('Sunday', 'sun '),
		('Monday', 'mon '),
		('Tuesday', 'tue '),
		('Wednesday', 'wed '),
		('Thursday', 'thu '),
		('Friday', 'fri '),
		('Saturday', 'sat '),
]


pseudotypes = [ # non-concrete types that are only used for documentation purposes; use to remap typesbycode
		('file', 'file'), # typically FileURL, but could be other file types as well
		('number', 'nmbr'), # any numerical type: Integer, Float, Long
		# ('text', 'ctxt'), # Word X, Excel X uses 'ctxt' instead of 'TEXT' or 'utxt' (TO CHECK: is this Excel's stupidity, or is it acceptable?)
]


properties = [
		('class_', 'pcls'), # used as a key in AERecord structures that have a custom class; also, Apple have removed the 'item' class definition from skeleton.sdef
		('properties', 'pALL'), # Apple have removed the 'item' class definition from skeleton.sdef
		('id', 'ID  '), # some apps (e.g. iTunes) may omit 'id' property from terminology despite using it
]


elements = [
		('items', 'cobj'), # Apple have removed the 'item' class definition from skeleton.sdef
]


enumerations = [
		('savo', [
				('yes', 'yes '), 
				('no', 'no  '), 
				('ask', 'ask '),
		]),
		# constants used in commands' 'ignore' argument (note: most apps currently ignore these):
		('cons', [
			('case', 'case'),
			('diacriticals', 'diac'),
			('expansion', 'expa'),
			('punctuation', 'punc'),
			('hyphens', 'hyph'),
			('whitespace', 'whit'),
			('numeric_strings', 'nume'),
			('application_responses', 'rmte'),
		]),
]


commands = [
	# required suite
	('run', 'aevtoapp', []),
	('open', 'aevtodoc', []),
	('print_', 'aevtpdoc', []),
	('quit', 'aevtquit', [('saving', 'savo')]),
	# 'reopen' and 'activate' aren't listed in required suite, but should be
	('reopen', 'aevtrapp', []),
	('activate', 'miscactv', []),
	# 'launch' is a special case not listed in the required suite and implementation is provided by
	# the Apple event bridge (not the target applications), which uses the Process Manager/
	# LaunchServices to launch an application without sending it the usual run/open event.
	('launch', 'ascrnoop', []),
	# 'get' and 'set' commands are often omitted from applications' core suites, even when used
	('get', 'coregetd', []),
	('set', 'coresetd', [('to', 'data')]),
	# some apps (e.g. Safari) which support GetURL events may omit it from their terminology; 
	# 'open location' is the name Standard Additions defines for this event, so use it here
	('open_location', 'GURLGURL', [('window', 'WIND')]), 
]


